import pyvista as pv
import numpy as np
from copy import deepcopy
import tqdm
import warnings
from itertools import permutations
from scipy.spatial import cKDTree
from scipy.sparse.csgraph import min_weight_full_bipartite_matching
from svv.utils.remeshing.remesh import remesh_surface_2d, remesh_volume
import tetgen
import vtk


class BoundaryLayer(object):
    def __init__(self, surface, layer_thickness=1.0, layer_thickness_ratio=0.5,
                 constant_thickness=True, max_layer_thickness=None, min_layer_thickness=None,
                 number_of_sublayers=3, number_of_substeps=100, sublayer_ratio=0.3,
                 relaxation=0.01, local_correction_factor=0.45, include_surface_cells=True,
                 include_sidewall_cells=True, negate_warp_vectors=True, volume_cell_entity_id=0,
                 inner_surface_cell_entity_id=1, outer_surface_cell_entity_id=2,
                 use_warp_vector_magnitude_as_thickness=False, layer_thickness_array=None,
                 warp_vector_array="Normals", remesh_vol=False, combine=True):
        self.surface = surface
        self.warp_vector_array = warp_vector_array
        self.surface_cells = {}
        self.surface_cell_array = self.surface.faces.reshape(-1, 4)[:, 1:]
        self.point_cells = {}
        for i in range(surface.n_points):
            self.point_cells[i] = []
        for i in tqdm.trange(surface.n_cells, desc="Processing Surface Cells", leave=False):
            cell = surface.get_cell(i)
            data = {}
            data['points'] = cell.points
            data['point_ids'] = cell.point_ids
            data['type'] = cell.type
            for j in cell.point_ids:
                self.point_cells[j].append(i)
            self.surface_cells[i] = data
        self.point_neighbors = {}
        for i in tqdm.trange(surface.n_points, desc="Processing Point Neighbors", leave=False):
            self.point_neighbors[i] = []
            for j in self.point_cells[i]:
                cell = self.surface_cells[j]
                for k in cell['point_ids']:
                    if k != i:
                        self.point_neighbors[i].append(k)
        if warp_vector_array not in surface.point_data.keys():
            if warp_vector_array == "Normals":
                print("Normals not found, computing normals.")
                self.surface.compute_normals(inplace=True)
            else:
                raise ValueError("Warp vector array not found.")
        self.layer_thickness = layer_thickness
        self.layer_thickness_ratio = layer_thickness_ratio
        self.constant_thickness = constant_thickness
        self.max_layer_thickness = max_layer_thickness
        self.min_layer_thickness = min_layer_thickness
        self.number_of_sublayers = number_of_sublayers
        self.number_of_substeps = number_of_substeps
        self.sublayer_ratio = sublayer_ratio
        self.relaxation = relaxation
        self.local_correction_factor = local_correction_factor
        self.include_surface_cells = include_surface_cells
        self.include_sidewall_cells = include_sidewall_cells
        self.negate_warp_vectors = negate_warp_vectors
        self.boundary_layers = pv.UnstructuredGrid()
        self.volume_cell_entity_id = volume_cell_entity_id
        self.inner_surface_cell_entity_id = inner_surface_cell_entity_id
        self.outer_surface_cell_entity_id = outer_surface_cell_entity_id
        self.use_warp_vector_magnitude_as_thickness = use_warp_vector_magnitude_as_thickness
        self.layer_thickness_array = layer_thickness_array
        self.warp_vectors = None
        self.remesh_vol = remesh_vol
        self.tolerance = 1E-14
        self.layers = None
        self.combine = combine

    def generate(self):
        grid = deepcopy(self.surface)
        boundary_layer_cell_array = []
        boundary_layer_celltype_array = []
        cell_entity_ids_array = []
        output_surface_cellids_array = []
        quadratic = False
        if quadratic:
            number_layer_points = 2 * grid.n_points
        else:
            number_layer_points = grid.n_points
        output_points = np.zeros((grid.n_points + number_layer_points * self.number_of_sublayers, 3))
        output_points[0:grid.n_points, :] = grid.points.astype(np.float64)
        if self.include_surface_cells:
            for i in range(grid.n_cells):
                #cell = grid.get_cell(i)
                cell = self.surface_cells[i]
                surfacePts = []
                if cell['type'].name == 'TRIANGLE':
                    boundary_layer_celltype_array.append(pv.CellType.TRIANGLE)
                    surfacePts.append(cell['point_ids'][0])
                    surfacePts.append(cell['point_ids'][1])
                    surfacePts.append(cell['point_ids'][2])
                elif cell['type'].name == 'QUAD':
                    boundary_layer_celltype_array.append(pv.CellType.QUAD)
                    surfacePts.append(cell['point_ids'][0])
                    surfacePts.append(cell['point_ids'][1])
                    surfacePts.append(cell['point_ids'][2])
                    surfacePts.append(cell['point_ids'][3])
                elif cell['type'].name == 'QUADRATIC_TRIANGLE':
                    pass
                else:
                    raise ValueError("Unsupported cell type.")
                surfacePts.insert(0, len(surfacePts))
                boundary_layer_cell_array.append(surfacePts)
                cell_entity_ids_array.append(self.inner_surface_cell_entity_id)
                output_surface_cellids_array.append(i)
        relaxation = self.relaxation
        initial_number_of_steps = int(self.number_of_substeps / 100)
        intermediate_number_of_steps = int(self.number_of_substeps / 10)
        final_number_of_steps = int(self.number_of_substeps - initial_number_of_steps - intermediate_number_of_steps)
        self.build_warp_vectors(grid)
        self.increment_warp_vectors(grid, initial_number_of_steps, relaxation)
        tangled_cell_ids = self.check_tangle(grid)
        maximum_iterations = int(self.number_of_substeps / 10)
        iteration = 0
        while len(tangled_cell_ids) > 0 and iteration < maximum_iterations:
            self.increment_warp_vectors(grid, intermediate_number_of_steps, relaxation)
            tangled_cell_ids = self.check_tangle(grid)
            self.local_untangle_fast(grid, tangled_cell_ids, self.local_correction_factor)
            tangled_cell_ids = self.check_tangle(grid)
            iteration += 1
            if len(tangled_cell_ids) == 0:
                print("try last : ", iteration)
                self.increment_warp_vectors(grid, final_number_of_steps, relaxation)
                tangled_cell_ids = self.check_tangle(grid)
        input_points = np.array(grid.points, dtype=np.float64)
        for i in tqdm.trange(self.number_of_sublayers, desc="Generating Boundary Layers", leave=False):
            warp_points = self.init_warp_points(input_points, i, False)
            if np.any(np.linalg.norm(warp_points - input_points, axis=1) == 0.0):
                print("Warp points not initialized correctly.")
            output_points[grid.n_points + i * number_layer_points:grid.n_points + (i + 1) * number_layer_points, :] = warp_points
            for j in range(grid.n_cells):
                #cell = grid.get_cell(j)
                cell = self.surface_cells[j]
                if cell['type'].name == 'TRIANGLE' or cell['type'].name == 'QUAD':
                    prismNPts = len(cell['point_ids']) * 2
                    prismPts = []
                    quadNPts = 4
                    quadPts = []
                    for k in range(len(cell['point_ids'])):
                        prismPts.append(cell['point_ids'][k] + i*number_layer_points)
                    for k in range(len(cell['point_ids'])):
                        prismPts.append(cell['point_ids'][k] + (i + 1) * number_layer_points)
                    prismPts.insert(0, prismNPts)
                    boundary_layer_cell_array.append(prismPts)
                    cell_entity_ids_array.append(self.volume_cell_entity_id)
                    output_surface_cellids_array.append(-1)
                    if cell['type'].name == 'TRIANGLE':
                        boundary_layer_celltype_array.append(pv.CellType.WEDGE)
                    elif cell['type'].name == 'QUAD':
                        boundary_layer_celltype_array.append(pv.CellType.HEXAHEDRON)
                    if self.include_sidewall_cells:
                        pass
                elif cell['type'].name == 'QUADRATIC_TRIANGLE':
                    if self.include_sidewall_cells:
                        pass
                    pass
                else:
                    raise ValueError("Unsupported cell type.")
            if self.include_surface_cells:
                if i == self.number_of_sublayers - 1:
                    for j in range(grid.n_cells):
                        #cell = grid.get_cell(j)
                        cell = self.surface_cells[j]
                        surfacePts = []
                        if cell['type'].name == 'TRIANGLE':
                            boundary_layer_celltype_array.append(pv.CellType.TRIANGLE)
                            surfacePts.append(cell['point_ids'][0] + i * number_layer_points + number_layer_points)
                            surfacePts.append(cell['point_ids'][1] + i * number_layer_points + number_layer_points)
                            surfacePts.append(cell['point_ids'][2] + i * number_layer_points + number_layer_points)
                        elif cell['type'].name == 'QUAD':
                            boundary_layer_celltype_array.append(pv.CellType.QUAD)
                            surfacePts.append(cell['point_ids'][0] + i * number_layer_points + number_layer_points)
                            surfacePts.append(cell['point_ids'][1] + i * number_layer_points + number_layer_points)
                            surfacePts.append(cell['point_ids'][2] + i * number_layer_points + number_layer_points)
                            surfacePts.append(cell['point_ids'][3] + i * number_layer_points + number_layer_points)
                        elif cell['type'].name == 'QUADRATIC_TRIANGLE':
                            pass
                        else:
                            raise ValueError("Unsupported cell type.")
                        surfacePts.insert(0, len(surfacePts))
                        boundary_layer_cell_array.append(surfacePts)
                        cell_entity_ids_array.append(self.outer_surface_cell_entity_id)
                        output_surface_cellids_array.append(j)
        all_cell_array = []
        for cell in boundary_layer_cell_array:
            all_cell_array.extend(cell)
        boundary_layer_cell_array = np.array(all_cell_array).ravel()
        layers = pv.UnstructuredGrid(boundary_layer_cell_array, boundary_layer_celltype_array, output_points)
        layers.cell_data['EntityID'] = np.array(cell_entity_ids_array)
        layers = self.tetra_filter(layers)
        self.layers = layers
        if self.combine:
            combined, interior = self.build_combined(layers, tolerance=self.tolerance)
        else:
            combined = None
            interior = None
        return combined, interior, layers

    def build_warp_vectors(self, grid):
        self.warp_vectors = np.array(grid.point_data[self.warp_vector_array], dtype=np.float64)
        if self.negate_warp_vectors:
            self.warp_vectors = -self.warp_vectors
        layer_thickness = 0.0
        if self.constant_thickness:
            layer_thickness = self.layer_thickness
        elif self.use_warp_vector_magnitude_as_thickness:
            layer_thickness = np.linalg.norm(self.warp_vectors, axis=1).reshape(-1, 1)
        else:
            layer_thickness = self.layer_thickness_array * self.layer_thickness_ratio
        if self.max_layer_thickness is not None:
            layer_thickness = np.minimum(layer_thickness, self.max_layer_thickness)
        magnitudes = np.linalg.norm(self.warp_vectors, axis=1).reshape(-1, 1)
        if np.any(magnitudes == 0.0):
            raise ValueError("Warp vectors have zero magnitude.")
        self.warp_vectors = (self.warp_vectors / magnitudes) * layer_thickness
        return

    def increment_warp_vectors(self, grid, number_of_substeps, relaxation):
        input_points = np.array(grid.points, dtype=np.float64)
        base_points = np.array(grid.points, dtype=np.float64)
        warp_points = np.zeros_like(base_points)
        for step in range(number_of_substeps):
            warp_points = self.inrcemental_warp_points(grid, base_points, warp_points, step, number_of_substeps, relaxation)
            base_points = deepcopy(warp_points)
        self.warp_vectors = warp_points - input_points
        layer_thickness = np.linalg.norm(self.warp_vectors, axis=1).reshape(-1, 1)
        self.warp_vectors = self.warp_vectors / layer_thickness
        self.warp_vectors = self.warp_vectors * layer_thickness
        return

    def inrcemental_warp_points(self, grid, base_points, warp_points, step, number_of_steps, relaxation):
        layer_thickness = np.linalg.norm(self.warp_vectors, axis=1).reshape(-1, 1)
        warp_vectors = self.warp_vectors / layer_thickness.reshape(-1, 1)
        layer_thickness /= number_of_steps
        warp_points = base_points + warp_vectors * layer_thickness
        edges = grid.extract_feature_edges(boundary_edges=True, manifold_edges=False,
                                                 feature_edges=False, non_manifold_edges=False)
        edge_pt_ids = []
        all_points = np.array(grid.points, dtype=np.float64)
        for i in range(edges.n_points):
            edge_pt_ids.append(np.argwhere(np.all(edges.points[i, :] == grid.points, axis=1)).flatten()[0])
        for i in range(grid.points.shape[0]):
            #neighbors = grid.point_neighbors(i)
            neighbors = self.point_neighbors[i]
            # Determine if point is on the edge of the surface
            if i in edge_pt_ids:
                continue
            barycenter = all_points[neighbors].mean(axis=0)
            warp_points[i, :] += relaxation * (barycenter - warp_points[i, :])
        return warp_points

    def check_tangle(self, grid):
        found = np.zeros(grid.n_cells)
        all_points = np.array(grid.points, dtype=np.float64)
        #for i in range(grid.n_cells):
        #    #cell = grid.get_cell(i)
        #    cell = self.surface_cells[i]
        #    cell_points = all_points[cell['point_ids'], :]
        #    warped_points = cell_points + self.warp_vectors[cell['point_ids'], :]
        #    base_normal = self.compute_normal(cell_points)
        #    warped_normal = self.compute_normal(warped_points)
        #    base_area = self.compute_area(cell_points)
        #    warped_area = self.compute_area(warped_points)
        #    if np.dot(base_normal, warped_normal) < 0.0 or warped_area / base_area < 0.1:
        #        found.append(i)
        cell_points = all_points[self.surface_cell_array, :]
        warped_points = cell_points + self.warp_vectors[self.surface_cell_array, :]
        base_normals = self.compute_normals(cell_points)
        warped_normals = self.compute_normals(warped_points)
        base_area = self.compute_areas(cell_points)
        warped_area = self.compute_areas(warped_points)
        normal_dots = (np.linalg.norm(base_normals * warped_normals, axis=1) < 0.0).flatten()
        found[normal_dots] = 1
        area_ratios = ((warped_area / base_area) <= 0.1).flatten()
        found[area_ratios] = 1
        found = np.argwhere(found == 1).flatten().tolist()
        return found

    def local_untangle(self, grid, tangled_cell_ids, alpha):
        correction_array = np.zeros((grid.n_points, 3), dtype=np.float64)
        for i in range(grid.n_cells):
            if i not in tangled_cell_ids:
                continue
            #cell = grid.get_cell(i)
            cell = self.surface_cells[i]
            warp_vectors = self.warp_vectors[cell['point_ids'], :]
            # SimVascular Implementation only does 1 iteration?
            for j in range(1):
                w1s = np.zeros(3)
                w2s = np.zeros(3)
                w3s = np.zeros(3)
                neighbors = grid.point_cell_ids(cell['point_ids'][0])
                for k in neighbors:
                    #neighbor_cell = grid.get_cell(k)
                    neighbor_cell = self.surface_cells[k]
                    n = self.compute_normal(neighbor_cell['points'])
                    w1s += warp_vectors[0, :] - np.dot(warp_vectors[0, :], n) * n
                w1s_mag = np.linalg.norm(w1s)
                w1s = w1s / w1s_mag
                neighbors = grid.point_cell_ids(cell['point_ids'][1])
                for k in neighbors:
                    #neighbor_cell = grid.get_cell(k)
                    neighbor_cell = self.surface_cells[k]
                    n = self.compute_normal(neighbor_cell['points'])
                    w2s += warp_vectors[1, :] - np.dot(warp_vectors[1, :], n) * n
                w2s_mag = np.linalg.norm(w2s)
                w2s = w2s / w2s_mag
                neighbors = grid.point_cell_ids(cell['point_ids'][2])
                for k in neighbors:
                    #neighbor_cell = grid.get_cell(k)
                    neighbor_cell = self.surface_cells[k]
                    n = self.compute_normal(neighbor_cell['points'])
                    w3s += warp_vectors[2, :] - np.dot(warp_vectors[2, :], n) * n
                w3s_mag = np.linalg.norm(w3s)
                w3s = w3s / w3s_mag
                correction_array[cell['point_ids'][0], :] += alpha * w2s + alpha * w3s
                correction_array[cell['point_ids'][1], :] += alpha * w1s + alpha * w3s
                correction_array[cell['point_ids'][2], :] += alpha * w1s + alpha * w2s
        layer_thickness = np.linalg.norm(self.warp_vectors, axis=1).reshape(-1, 1)
        self.warp_vectors = (self.warp_vectors + correction_array)
        self.warp_vectors = self.warp_vectors / np.linalg.norm(self.warp_vectors, axis=1).reshape(-1, 1)
        self.warp_vectors = self.warp_vectors * layer_thickness
        if np.any(layer_thickness == 0.0):
            warnings.warn("Warp vectors have zero magnitude.")
        return

    def local_untangle_fast(self, grid, tangled_cell_ids, alpha):
        correction_array = np.zeros((grid.n_points, 3), dtype=np.float64)
        all_cells = self.surface.faces.reshape(-1, 4)[:, 1:]

        for i in tqdm.trange(grid.n_cells, desc="Local Untangle", leave=False):
            if i not in tangled_cell_ids:
                continue
            #cell = grid.get_cell(i)
            cell = self.surface_cells[i]
            warp_vectors = self.warp_vectors[cell['point_ids'], :]
            # SimVascular Implementation only does 1 iteration?
            for j in range(1):
                w1s = np.zeros(3)
                w2s = np.zeros(3)
                w3s = np.zeros(3)
                #neighbors = grid.point_cell_ids(cell['point_ids'][0])
                neighbors = self.point_cells[cell['point_ids'][0]]
                for k in neighbors:
                    #neighbor_cell = grid.get_cell(k)
                    neighbor_cell = self.surface_cells[k]
                    n = self.compute_normal(neighbor_cell['points'])
                    w1s += warp_vectors[0, :] - np.dot(warp_vectors[0, :], n) * n
                w1s_mag = np.linalg.norm(w1s)
                w1s = w1s / w1s_mag
                #neighbors = grid.point_cell_ids(cell['point_ids'][1])
                neighbors = self.point_cells[cell['point_ids'][1]]
                for k in neighbors:
                    #neighbor_cell = grid.get_cell(k)
                    neighbor_cell = self.surface_cells[k]
                    n = self.compute_normal(neighbor_cell['points'])
                    w2s += warp_vectors[1, :] - np.dot(warp_vectors[1, :], n) * n
                w2s_mag = np.linalg.norm(w2s)
                w2s = w2s / w2s_mag
                #neighbors = grid.point_cell_ids(cell['point_ids'][2])
                neighbors = self.point_cells[cell['point_ids'][2]]
                for k in neighbors:
                    #neighbor_cell = grid.get_cell(k)
                    neighbor_cell = self.surface_cells[k]
                    n = self.compute_normal(neighbor_cell['points'])
                    w3s += warp_vectors[2, :] - np.dot(warp_vectors[2, :], n) * n
                w3s_mag = np.linalg.norm(w3s)
                w3s = w3s / w3s_mag
                correction_array[cell['point_ids'][0], :] += alpha * w2s + alpha * w3s
                correction_array[cell['point_ids'][1], :] += alpha * w1s + alpha * w3s
                correction_array[cell['point_ids'][2], :] += alpha * w1s + alpha * w2s
        layer_thickness = np.linalg.norm(self.warp_vectors, axis=1).reshape(-1, 1)
        self.warp_vectors = (self.warp_vectors + correction_array)
        self.warp_vectors = self.warp_vectors / np.linalg.norm(self.warp_vectors, axis=1).reshape(-1, 1)
        self.warp_vectors = self.warp_vectors * layer_thickness
        if np.any(layer_thickness == 0.0):
            warnings.warn("Warp vectors have zero magnitude.")
        return

    def init_warp_points(self, input_points, sub_layer_id, quadratic):
        total_layer_zero_sublayer_ratio = 0.0
        for i in range(self.number_of_sublayers):
            total_layer_zero_sublayer_ratio += self.sublayer_ratio ** (self.number_of_sublayers - i - 1)
        sublayer_offset_ratio = 0.0
        for i in range(0, sub_layer_id):
            sublayer_offset_ratio += (self.sublayer_ratio ** (self.number_of_sublayers - i - 1))
        sublayer_offset_ratio /= total_layer_zero_sublayer_ratio
        sublayer_thickness_ratio = (self.sublayer_ratio ** (self.number_of_sublayers - sub_layer_id - 1)) / total_layer_zero_sublayer_ratio
        if not quadratic:
            warped_points = np.zeros((input_points.shape[0], 3))
            layer_thickness = np.linalg.norm(self.warp_vectors, axis=1).reshape(-1, 1)
            warp_vectors = self.warp_vectors / layer_thickness
            sublayer_offset = sublayer_offset_ratio * layer_thickness
            sublayer_thickness = sublayer_thickness_ratio * layer_thickness
            warped_points[:, :] = input_points + warp_vectors * (sublayer_offset + sublayer_thickness)
        else:
            warped_points = np.zeros((2 * input_points.shape[0], 3))
            layer_thickness = np.linalg.norm(self.warp_vectors, axis=1).reshape(-1, 1)
            warp_vectors = self.warp_vectors / layer_thickness
            sublayer_offset = sublayer_offset_ratio * layer_thickness
            sublayer_thickness = sublayer_thickness_ratio * layer_thickness
            warped_points[0:input_points.shape[0], :] = input_points + 0.5 * warp_vectors * (sublayer_offset + sublayer_thickness)
            warped_points[input_points.shape[0]:, :] = input_points + warp_vectors * (sublayer_offset + sublayer_thickness)
        return warped_points

    def compute_normal(self, pts):
        n = np.cross(pts[1] - pts[0], pts[2] - pts[0])
        return n / np.linalg.norm(n)

    def compute_area(self, pts):
        n = np.cross(pts[1] - pts[0], pts[2] - pts[0])
        return 0.5 * np.linalg.norm(n)

    def compute_normals(self, pts):
        n = np.zeros((pts.shape[0], 3), dtype=np.float64)
        n[:, :] = np.cross(pts[:, 1, :] - pts[:, 0, :], pts[:, 2, :] - pts[:, 0, :])
        return n

    def compute_areas(self, pts):
        n = np.zeros((pts.shape[0], 3), dtype=np.float64)
        n[:, :] = np.cross(pts[:, 1, :] - pts[:, 0, :], pts[:, 2, :] - pts[:, 0, :])
        return 0.5 * np.linalg.norm(n, axis=1)

    def build_connections(self, layers):
        connections = {'vol_to_vol':{}, 'vol_to_faces':{}, 'faces_to_vol':{}, "faces_to_faces":{}}
        volume_cells = np.argwhere(layers.cell_data['EntityID'] == self.volume_cell_entity_id).flatten()
        volume_cells = layers.extract_cells(volume_cells)

    def tetra_filter(self, layers):
        """
        This function filters the boundary layer mesh to yield a tetrahedral mesh.
        We assume that the volume elements of the unstructured grid are well-posed
        in that they define a closed volume and represent a valid element template.
        However, we do not make assumptions regarding orientation of the volume
        elements.
        :param layers:
        :return:
        """
        tetra_ordering = list(permutations([1, 2, 3, 4]))
        inner_submesh_cells = []
        inner_submesh_cells_type = []
        inner_submesh_cells_entity = []
        outer_submesh_cells = []
        outer_submesh_cells_type = []
        outer_submesh_cells_entity = []
        inner_submesh_ids = np.argwhere(layers.cell_data['EntityID'] == self.inner_surface_cell_entity_id).flatten()
        outer_submesh_ids = np.argwhere(layers.cell_data['EntityID'] == self.outer_surface_cell_entity_id).flatten()
        for i in inner_submesh_ids:
            cell = layers.get_cell(i)
            nodes = cell.point_ids
            nodes.insert(0, len(nodes))
            inner_submesh_cells.append(nodes)
            inner_submesh_cells_type.append(cell.type)
            inner_submesh_cells_entity.append(self.inner_surface_cell_entity_id)
        for i in outer_submesh_ids:
            cell = layers.get_cell(i)
            nodes = cell.point_ids
            nodes.insert(0, len(nodes))
            outer_submesh_cells.append(nodes)
            outer_submesh_cells_type.append(cell.type)
            outer_submesh_cells_entity.append(self.outer_surface_cell_entity_id)
        volume_submesh = layers.extract_cells(np.argwhere(layers.cell_data['EntityID'] == self.volume_cell_entity_id).flatten())
        volume_submesh = volume_submesh.triangulate(progress_bar=True)
        volume_submesh_quality = volume_submesh.compute_cell_quality().cell_data['CellQuality'].flatten()
        #if np.any(volume_submesh_quality < 0):
        #    old_tetra_cells = deepcopy(volume_submesh.cells.reshape(-1, 5))
        #    new_tetra_cells = deepcopy(volume_submesh.cells.reshape(-1, 5))
        #    tetra_cell_type = [pv.CellType.TETRA for i in range(new_tetra_cells.shape[0])]
        #    output_points = np.array(volume_submesh.points)
        #    for i in tqdm.trange(len(tetra_ordering), desc="Reordering Tetrahedra", leave=False):
        #        ordering = tetra_ordering[i]
        #        ids = np.argwhere(volume_submesh_quality < 0)
        #        #changed = tetra_cells[ids, ordering].copy()
        #        new_tetra_cells[ids.flatten(), 1:] = old_tetra_cells[ids, ordering]
        #        new_volume_submesh = pv.UnstructuredGrid(new_tetra_cells.ravel(), tetra_cell_type, output_points)
        #        quality = new_volume_submesh.compute_cell_quality().cell_data['CellQuality'].flatten()
        #        if not np.any(quality < 0):
        #            break
        volume_submesh_cells = volume_submesh.cells.reshape(-1, 5).tolist()
        volume_submesh_cells_type = [pv.CellType.TETRA] * len(volume_submesh_cells)
        volume_submesh_entity = [self.volume_cell_entity_id] * len(volume_submesh_cells)
        all_cells = []
        all_cells.extend(inner_submesh_cells)
        all_cells.extend(volume_submesh_cells)
        all_cells.extend(outer_submesh_cells)
        all_cells_type = []
        all_cells_type.extend(inner_submesh_cells_type)
        all_cells_type.extend(volume_submesh_cells_type)
        all_cells_type.extend(outer_submesh_cells_type)
        all_cells_entity = []
        all_cells_entity.extend(inner_submesh_cells_entity)
        all_cells_entity.extend(volume_submesh_entity)
        all_cells_entity.extend(outer_submesh_cells_entity)
        tmp = []
        for i in range(len(all_cells)):
            tmp.extend(all_cells[i])
        all_cells = np.array(tmp).ravel()
        all_cells_type = np.array(all_cells_type)
        all_cells_entity = np.array(all_cells_entity)
        all_points = np.array(layers.points)
        tetra_mesh = pv.UnstructuredGrid(all_cells, all_cells_type, all_points)
        tetra_mesh.cell_data['EntityID'] = all_cells_entity
        return tetra_mesh

    def build_combined(self, layers, tolerance=1E-14):
        # The points of the mesh surface need to be shifted to correct the mesh for the
        # inclusion of the boundary layers
        layers = deepcopy(layers)
        inner_surface_cells = np.argwhere(layers.cell_data['EntityID'] == self.inner_surface_cell_entity_id).flatten()
        outer_surface_cells = np.argwhere(layers.cell_data['EntityID'] == self.outer_surface_cell_entity_id).flatten()
        inner_surface_pt_ids = []
        outer_surface_pt_ids = []
        for i in tqdm.trange(len(inner_surface_cells), desc="Collecting Inner Surface Points", leave=False):
            cell_id = inner_surface_cells[i]
            cell = layers.get_cell(cell_id)
            inner_surface_pt_ids.extend(cell.point_ids)
        for i in tqdm.trange(len(outer_surface_cells), desc="Collecting Outer Surface Points", leave=False):
            cell_id = outer_surface_cells[i]
            cell = layers.get_cell(cell_id)
            outer_surface_pt_ids.extend(cell.point_ids)
        outer = layers.extract_cells(outer_surface_cells)
        layers_tree = cKDTree(layers.points)
        dists, inds = layers_tree.query(outer.points, k=2)
        if np.any(dists[:, 1] == 0.0):
            warnings.warn("Internal Layer Points coincident with Outer Surface Points.")
        original_point_ids = outer.point_data['vtkOriginalPointIds']
        outer = outer.extract_surface()
        outer_tree = cKDTree(outer.points)

        boundaries = outer.extract_feature_edges(boundary_edges=True, manifold_edges=False, feature_edges=False,
                                                 non_manifold_edges=False)
        boundaries = boundaries.split_bodies()
        caps = []
        for i in range(len(boundaries)):
            cap = remesh_surface_2d(boundaries[i])
            #_, outer_cap_inds = outer_tree.query(cap.points)
            #cap.points[:boundaries[i].n_points, :] = outer.points[outer_cap_inds[:boundaries[i].n_points], :]
            caps.append(cap)
        caps.insert(0, outer)
        outer_total = pv.merge(caps, merge_points=True, progress_bar=True)
        outer_total = outer_total.clean(tolerance=tolerance, progress_bar=True)
        if not outer_total.is_manifold:
            raise ValueError("Merged Outer Surface is not manifold.")
        else:
            print("Merged Outer Surface is manifold.")
        outer_total_quality = outer_total.compute_cell_quality().cell_data["CellQuality"]
        if np.any(outer_total_quality < 0.0):
            warnings.warn("Outer Surface has inverted triangles.")
        interior = tetgen.TetGen(outer_total)
        interior.tetrahedralize(switches='pq1.1/10MVYSJ')
        mesh_interior = interior.grid
        if self.remesh_vol:
            print("Remeshing interior volume...")
            mesh_interior = remesh_volume(mesh_interior, nosurf=True)
            print("Remeshing interior complete.")
        mesh_interior = mesh_interior.triangulate().clean()
        mesh_interior_surface = mesh_interior.extract_surface()
        if not mesh_interior_surface.is_manifold:
            warnings.warn("Interior Surface is not manifold.")
        else:
            print("Interior Surface is manifold.")
        layers_volume = layers.extract_cells(np.argwhere(layers.cell_data['EntityID'] == self.volume_cell_entity_id).flatten())
        layers_volume = layers_volume.triangulate().clean()
        layers_surface = layers_volume.extract_surface()
        if not layers_surface.is_manifold:
            warnings.warn("Layers Surface is not manifold.")
        else:
            print("Layers Surface is manifold.")
        # Check Quality of Combined Mesh
        quality = mesh_interior.compute_cell_quality().cell_data['CellQuality'].flatten()
        if np.any(quality < 0):
            warnings.warn("Combined Mesh has negative jacobian at {}".format(np.argwhere(quality < 0).flatten()))
            print("Attempting to fix negative jacobians via volume meshing.")
        """
        layer_to_outer_dists, layer_to_outer_inds = outer_tree.query(layers_volume.points)
        outer_to_layer = {}
        tolerance = 1E-6
        for i in tqdm.trange(len(layer_to_outer_inds), desc="Mapping Layer Points to Outer Surface", leave=False):
            if layer_to_outer_dists[i] < tolerance:
                if outer_to_layer.get(layer_to_outer_inds[i], None) is None:
                    outer_to_layer[layer_to_outer_inds[i]] = i
                else:
                    warnings.warn("Multiple Layer Points Mapped to Single Outer Point.")
        interior_to_outer_dists, interior_to_outer_inds = outer_tree.query(mesh_interior.points)
        outer_to_interior = {}
        for i in tqdm.trange(len(interior_to_outer_inds), desc="Mapping Interior Points to Outer Surface", leave=False):
            if interior_to_outer_dists[i] < tolerance:
                if outer_to_interior.get(interior_to_outer_inds[i], None) is None:
                    outer_to_interior[interior_to_outer_inds[i]] = i
                else:
                    warnings.warn("Multiple Interior Points Mapped to Single Outer Point.")
        layer_cells = layers_volume.cells.reshape(-1, 5)[:, 1:]
        """

        appender = vtk.vtkAppendDataSets()
        appender.AddInputData(layers_volume)
        appender.AddInputData(mesh_interior)
        appender.SetTolerance(tolerance) #need to change from hard-coded value
        appender.MergePointsOn()
        appender.Update()
        combined_mesh = pv.wrap(appender.GetOutput())
        #combined_mesh = combined_mesh.clean()
        #combined_mesh = pv.merge([layers_volume, mesh_interior])
        #combined_mesh = combined_mesh.triangulate().clean()
        combined_mesh_surface = combined_mesh.extract_surface()
        if not combined_mesh_surface.is_manifold:
            non_manifold_edges = combined_mesh_surface.extract_feature_edges(boundary_edges=False, manifold_edges=False,
                                                                             feature_edges=False, non_manifold_edges=True)
            warnings.warn("Combined Mesh Surface is not manifold.\n {} Non-Manifold Edges at: {}".format(non_manifold_edges.n_cells, non_manifold_edges.cell_data["vtkOriginalCellIds"]) )
        else:
            print("Combined Mesh Surface is manifold.")
        # combined_mesh = pv.merge([layers_volume, mesh_interior], merge_points=True)
        # Replace the pv.merge function with a custom merge function
        return combined_mesh, mesh_interior

def append_boundary_layers(mesh, layers, inner_surface_cell_entity_id=1, outer_surface_cell_entity_id=2,
                           compute_inner_outer_mapping=False, scale=20):
    # The points of the mesh surface need to be shifted to correct the mesh for the
    # inclusion of the boundary layers
    mesh = deepcopy(mesh)
    layers = deepcopy(layers)
    inner_surface_cells = np.argwhere(layers.cell_data['EntityID'] == inner_surface_cell_entity_id).flatten()
    outer_surface_cells = np.argwhere(layers.cell_data['EntityID'] == outer_surface_cell_entity_id).flatten()
    inner_surface_pt_ids = []
    outer_surface_pt_ids = []
    for i in tqdm.trange(len(inner_surface_cells), desc="Collecting Inner Surface Points", leave=False):
        cell_id = inner_surface_cells[i]
        cell = layers.get_cell(cell_id)
        inner_surface_pt_ids.extend(cell.point_ids)
    for i in tqdm.trange(len(outer_surface_cells), desc="Collecting Outer Surface Points", leave=False):
        cell_id = outer_surface_cells[i]
        cell = layers.get_cell(cell_id)
        outer_surface_pt_ids.extend(cell.point_ids)
    inner_surface_pt_ids = np.unique(np.array(inner_surface_pt_ids).flatten())
    outer_surface_pt_ids = np.unique(np.array(outer_surface_pt_ids).flatten())
    inner_tree = cKDTree(layers.points[inner_surface_pt_ids])
    outer_tree = cKDTree(layers.points[outer_surface_pt_ids])
    mesh_tree = cKDTree(mesh.points)
    if compute_inner_outer_mapping:
        # Map the inner surface points to the outer surface points
        cell_volumes = mesh.compute_cell_sizes().cell_data['Volume']
        length = (cell_volumes.mean()**(1/3)) * scale
        sparse = inner_tree.sparse_distance_matrix(outer_tree, length, output_type='coo_matrix')
        row, col = min_weight_full_bipartite_matching(sparse)
        inner_to_outer_idx = outer_surface_pt_ids[col]
    else:
        inner_to_outer_idx = deepcopy(outer_surface_pt_ids)
    # Check that inner <-> outer mapping is 1:1
    _, counts = np.unique(inner_to_outer_idx, return_counts=True)
    if np.any(counts > 1):
        raise ValueError("Inner to Outer Surface Point Mapping is not 1:1.")
    # Map the inner surface points to the mesh surface points
    inner_to_mesh_dist, inner_to_mesh_idx = mesh_tree.query(layers.points[inner_surface_pt_ids])
    # Check that inner <-> mesh mapping is 1:1
    _, _, counts = np.unique(inner_to_mesh_idx, return_inverse=True, return_counts=True)
    if np.any(counts > 1):
        raise ValueError("Inner to Mesh Surface Point Mapping is not 1:1.")
    # Check that inner <-> mesh mapping is conforming
    if np.any(~np.isclose(inner_to_mesh_dist, 0.0)):
        raise ValueError("Inner to Mesh Surface Point Mapping is not conforming.\n Could result in mesh corruption.")
    outer = layers.extract_cells(outer_surface_cells).extract_surface()
    outer_tree = cKDTree(outer.points)
    boundaries = outer.extract_feature_edges(boundary_edges=True,manifold_edges=False,feature_edges=False,non_manifold_edges=False)
    boundaries = boundaries.split_bodies()
    caps = []
    for i in range(len(boundaries)):
        cap = remesh_surface_2d(boundaries[i])
        _, outer_cap_inds = outer_tree.query(cap.points)
        cap.points[:boundaries[i].n_points, :] = outer.points[outer_cap_inds[:boundaries[i].n_points], :]
        caps.append(cap)
    caps.insert(0, outer)
    outer_total = pv.merge(caps)
    interior = tetgen.TetGen(outer_total)
    interior.tetrahedralize()
    mesh = interior.grid
    #mesh.points[inner_to_mesh_idx[inner_surface_pt_ids]] = layers.points[inner_to_outer_idx]
    # Check Quality of Combined Mesh
    quality = mesh.compute_cell_quality().cell_data['CellQuality'].flatten()
    if np.any(quality < 0):
        warnings.warn("Combined Mesh has negative jacobian at {}".format(np.argwhere(quality < 0).flatten()))
        print("Attempting to fix negative jacobians via volume meshing.")
    #combined_mesh = pv.merge([layers, mesh], merge_points=True)
    appender = vtk.vtkAppendDataSets()
    appender.AddInputData(layers)
    appender.AddInputData(mesh)
    appender.Update()
    combined_mesh = pv.wrap(appender.GetOutput())
    return combined_mesh, mesh



from . import centerlines
from . import generate_1d_mesh
from . import io_0d
from . import io_1d
from . import io_headers
from . import manage
from . import mesh
from . import models
from . import parameters
from . import utils



# cython: language_level=3
# distutils: language=c++
import numpy as np
cimport numpy as cnp
import cython
from libc.math cimport sqrt, pow
from libcpp.vector cimport vector
from libcpp.unordered_map cimport unordered_map
from libcpp.unordered_set cimport unordered_set
from libc.math cimport acos, M_PI

cnp.import_array()

@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
def norm(double[:,:] data):
    cdef Py_ssize_t i = data.shape[0]
    cdef Py_ssize_t j = data.shape[1]
    cdef Py_ssize_t ii
    cdef cnp.ndarray[cnp.float64_t, ndim=2] magnitudes = np.zeros((i,1), dtype=np.float64)
    cdef double[:,:] mv = magnitudes
    cdef double mag
    for ii in range(i):
        mag = 0
        for jj in range(j):
            mag += pow(data[ii,jj], 2)
        mv[ii, 0] = sqrt(mag)
    return magnitudes

@cython.wraparound(False)
@cython.boundscheck(False)
@cython.nonecheck(False)
def argwhere_nonzeros(double[:] data):
    cdef Py_ssize_t i = data.shape[0]
    cdef Py_ssize_t ii
    cdef vector[int] nonzeros
    for ii in range(i):
        if data[ii] != 0:
            nonzeros.push_back(ii)
    cdef cnp.ndarray[cnp.int64_t, ndim=1] results = np.array(nonzeros, dtype=np.int64)
    return results

@cython.wraparound(False)
@cython.boundscheck(False)
@cython.nonecheck(False)
def argwhere_value_double(double[:] data, double value):
    cdef Py_ssize_t i = data.shape[0]
    cdef Py_ssize_t ii
    cdef vector[int] zeros
    for ii in range(i):
        if data[ii] == value:
            zeros.push_back(ii)
    cdef cnp.ndarray[cnp.int64_t, ndim=1] results = np.array(zeros, dtype=np.int64)
    return results

@cython.wraparound(False)
@cython.boundscheck(False)
@cython.nonecheck(False)
def argwhere_value_int(long long[:] data, long long value):
    cdef Py_ssize_t i = data.shape[0]
    cdef Py_ssize_t ii
    cdef vector[int] zeros
    for ii in range(i):
        if data[ii] == value:
            zeros.push_back(ii)
    cdef cnp.ndarray[cnp.int64_t, ndim=1] results = np.array(zeros, dtype=np.int64)
    return results


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.nonecheck(False)
def any_value_double(double[:] data, double value):
    cdef Py_ssize_t i = data.shape[0]
    cdef Py_ssize_t ii
    for ii in range(i):
        if data[ii] == value:
            return True
    return False


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.nonecheck(False)
def any_value_int(int[:] data, int value):
    cdef Py_ssize_t i = data.shape[0]
    cdef Py_ssize_t ii
    for ii in range(i):
        if data[ii] == value:
            return True
    return False


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.nonecheck(False)
def c_dict(double[:,:] data):
    cdef Py_ssize_t i = data.shape[0]
    cdef Py_ssize_t j = data.shape[1]
    cdef Py_ssize_t ii, jj
    cdef unordered_map[int, vector[double]] dictionary
    cdef int key
    cdef vector[double] value
    for ii in range(i):
        key = ii
        for jj in range(j):
            value.push_back(data[ii,jj])
        dictionary[key] = value
        value.clear()
    return dictionary


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.nonecheck(False)
def duplicate_map(long long[:] unique_inverse, long long[:] unique_counts):
    cdef Py_ssize_t i = unique_counts.shape[0]
    cdef Py_ssize_t j = unique_inverse.shape[0]
    cdef Py_ssize_t ii, jj
    cdef unordered_map[int, vector[int]] duplicate_dict
    cdef unordered_set[int] duplicate_set
    cdef int key
    cdef vector[int] value
    for ii in range(i):
        if unique_counts[ii] <= 1:
            continue
        else:
            for jj in range(j):
                if unique_inverse[jj] == ii:
                    value.push_back(jj)
                    duplicate_set.insert(jj)
        duplicate_dict[ii] = value
        value.clear()
    return duplicate_dict, duplicate_set

@cython.wraparound(False)
@cython.boundscheck(False)
@cython.nonecheck(False)
def _allocate_patch(long long[:] indices, double overlap, unordered_set[int] point_set, unordered_set[int] duplicates_set):
    """"
    Internal helper function for patch point allocation.
    This function takes a neighborhood points, the
    current remaining set of points, the set of
    duplicate points, and the overlap ratio
    and returns the set of points to be removed.
    """
    cdef Py_ssize_t i = indices.shape[0]
    cdef Py_ssize_t overlap_indicies = <Py_ssize_t>(i * overlap)
    for i in range(1, overlap_indicies):
        if point_set.count(indices[i]):
            if duplicates_set.size() > 0:
                if not duplicates_set.count(indices[i]):
                    point_set.erase(indices[i])
            else:
                point_set.erase(indices[i])
    return point_set

@cython.wraparound(False)
@cython.boundscheck(False)
@cython.nonecheck(False)
def closest_point(long long index, unordered_set[int] included, double[:,:] points):
    cdef double distance
    cdef double min_distance = 0.0
    cdef int closest_index
    cdef Py_ssize_t i = points.shape[0]
    cdef Py_ssize_t j = points.shape[1]
    for ii in included:
        distance = 0.0
        for jj in range(j):
            distance += pow(points[index, jj] - points[ii, jj], 2)
        distance = sqrt(distance)
        if distance > min_distance and min_distance == 0.0:
            min_distance = distance
            closest_index = ii
        elif distance == min_distance:
            closest_index = ii
        elif distance > min_distance:
            continue
        else:
            min_distance = distance
            closest_index = ii
    return closest_index, min_distance

@cython.wraparound(False)
@cython.boundscheck(False)
@cython.nonecheck(False)
def get_angle(int index_1, int index_2, double[:,:] normals):
    cdef double angle = 0.0
    cdef Py_ssize_t j = normals.shape[1]
    cdef Py_ssize_t jj
    for jj in range(j):
        angle = angle + normals[index_1, jj] * normals[index_2, jj]
    angle = acos(angle) * (180 / M_PI)
    return angle

@cython.wraparound(False)
@cython.boundscheck(False)
@cython.nonecheck(False)
def _allocate_angle(long long idx, long long[:] indices, double[:,:] points, double[:,:] normals, double feature_angle):
    cdef Py_ssize_t i = points.shape[0]
    cdef Py_ssize_t j = points.shape[1]
    cdef unordered_set[int] include_indices
    cdef unordered_set[int] include_points
    cdef vector[int] allocated_indices
    for ii in range(i):
        if indices[ii] == idx:
            include_points.insert(ii)
            include_indices.insert(indices[ii])
            allocated_indices.push_back(indices[ii])
    cdef int closest_idx
    cdef double closest_distance
    for ii in range(1, i):
        closest_idx, closest_distance = closest_point(ii, include_points, points)
        if closest_distance == 0.0:
            continue
        angle = get_angle(ii, closest_idx, normals)
        if angle > feature_angle:
            continue
        else:
            include_points.insert(ii)
            include_indices.insert(indices[ii])
            allocated_indices.push_back(indices[ii])
    return allocated_indices


import numpy
import numpy as np
from tqdm import trange

from geomdl import BSpline, utilities
from scipy import optimize
from scipy.interpolate import splprep, splev, CubicSpline
import pyvista as pv
from copy import deepcopy
from time import perf_counter
from svv.utils.spatial.c_distance import minimum_segment_distance
#from svv.forest.utils.angles import cost_angles
from svv.forest.connect.assign import assign_network, assign_network_vector

import matplotlib.pyplot as plt
from matplotlib.widgets import Slider
from mpl_toolkits.mplot3d import Axes3D
def test_vessels(size):
    vessels = numpy.random.random((size, 7))
    vessels[:, :6] *= 10
    return vessels


def test_lines(size):
    lines = numpy.zeros((size, 3))
    lines[:, 0] = numpy.linspace(0, 1, size) * 10
    lines[:, 1] = 0.5 * 10
    lines[:, 2] = 0.5 * 10
    return lines

class Spline(object):
    def __init__(self, pts, degree=3, cubic_hermite=True, weights=None, tangents=None):
        self.ctrlpts = pts
        self.degree = degree
        self.sample_size = None
        self.cubic_hermite = cubic_hermite
        if cubic_hermite:
            bc_num = 0
            if not isinstance(tangents, type(None)):
                if not isinstance(tangents[0], type(None)):
                    tangent_start = tangents[0]
                    bc_num += 1
                if not isinstance(tangents[1], type(None)):
                    tangent_end = tangents[1]
                    bc_num += 1
            else:
                tangent_start = None
                tangent_end = None
            if not isinstance(weights, type(None)):
                if not isinstance(weights[0], type(None)):
                    tangent_start = tangent_start * weights[0]
                else:
                    tangent_start = None
                if not isinstance(weights[1], type(None)):
                    tangent_end = tangent_end * weights[1]
                else:
                    tangent_end = None
            t = numpy.linspace(0, 1, pts.shape[0] - bc_num)
            tmp = numpy.zeros((pts.shape[0]-bc_num, 3))
            if bc_num == 2:
                tmp[0, :] = pts[0, :]
                tmp[-1, :] = pts[-1, :]
                tmp[1:-1, :] = pts[2:-2, :]
                self.spline_x = CubicSpline(t, tmp[:, 0], bc_type=((1, tangent_start[0]), (1, tangent_end[0])))
                self.spline_y = CubicSpline(t, tmp[:, 1], bc_type=((1, tangent_start[1]), (1, tangent_end[1])))
                self.spline_z = CubicSpline(t, tmp[:, 2], bc_type=((1, tangent_start[2]), (1, tangent_end[2])))
            elif bc_num == 1:
                if not isinstance(tangent_start, type(None)):
                    tmp[0, :] = pts[0, :]
                    tmp[1:, :] = pts[2:, :]
                    self.spline_x = CubicSpline(t, tmp[:, 0], bc_type=((1, tangent_start[0]), 'not-a-knot'))
                    self.spline_y = CubicSpline(t, tmp[:, 1], bc_type=((1, tangent_start[1]), 'not-a-knot'))
                    self.spline_z = CubicSpline(t, tmp[:, 2], bc_type=((1, tangent_start[2]), 'not-a-knot'))
                else:
                    tmp[-1, :] = pts[-1, :]
                    tmp[:-1, :] = pts[:-2, :]
                    self.spline_x = CubicSpline(t, tmp[:, 0], bc_type=('not-a-knot', (1, tangent_end[0])))
                    self.spline_y = CubicSpline(t, tmp[:, 1], bc_type=('not-a-knot', (1, tangent_end[1])))
                    self.spline_z = CubicSpline(t, tmp[:, 2], bc_type=('not-a-knot', (1, tangent_end[2])))
            else:
                tmp[:, :] = pts[:, :]
                self.spline_x = CubicSpline(t, tmp[:, 0])
                self.spline_y = CubicSpline(t, tmp[:, 1])
                self.spline_z = CubicSpline(t, tmp[:, 2])
        else:
            self.tck = splprep(pts.T, s=0, k=degree)
        self.evalpts = None

    def evaluate(self):
        t = numpy.linspace(0, 1, self.sample_size)
        if self.cubic_hermite:
            self.evalpts = numpy.vstack((self.spline_x(t), self.spline_y(t), self.spline_z(t))).T
        else:
            self.evalpts = numpy.array(splev(t, self.tck[0])).T
        return

    def __call__(self, t):
        if self.cubic_hermite:
            return numpy.vstack((self.spline_x(t), self.spline_y(t), self.spline_z(t))).T
        else:
            return numpy.array(splev(t, self.tck[0])).T

    def derivatives(self, order):
        t = numpy.linspace(0, 1, self.sample_size)
        der = None
        if self.cubic_hermite:
            dx = self.spline_x.derivative(order)
            dy = self.spline_y.derivative(order)
            dz = self.spline_z.derivative(order)
            der = numpy.vstack((dx(t), dy(t), dz(t)))
        else:
            der = numpy.array(splev(t, self.tck[0], der=order)).T
        return der

    def curvature(self, num):
        t_start = numpy.linspace(0, 0.1, num//4)[:-1]
        t = numpy.linspace(0.1, 0.9, num//2)
        t_end = numpy.linspace(0.9, 1, num//4)[1:]
        t = numpy.hstack((t_start, t, t_end))[1:-1]
        dx = self.spline_x.derivative(1)
        dy = self.spline_y.derivative(1)
        dz = self.spline_z.derivative(1)
        ddx = self.spline_x.derivative(2)
        ddy = self.spline_y.derivative(2)
        ddz = self.spline_z.derivative(2)
        curvature = (numpy.linalg.norm(numpy.cross(numpy.vstack((dx(t), dy(t), dz(t))), numpy.vstack((ddx(t), ddy(t), ddz(t))), axis=0), axis=0)
                                       / numpy.linalg.norm(numpy.vstack((dx(t), dy(t), dz(t))), axis=0)**3)
        return curvature


def spline_factory(p1, p2, p3, p4, clamp_first=True, clamp_second=True, ctlpts_func=None):
    def create_spline(data, p1_=p1, p2_=p2, p3_=p3, p4_=p4, c1=clamp_first, c2=clamp_second, ctlpts_func=ctlpts_func):
        """
        if c1 and c2:
            diff = data[1:-1].reshape(-1, 3)
            t = numpy.linspace(0, 1, diff.shape[0] + 2)
            pts = ctlpts_func(t)
            tmp0 = 0.5 * (numpy.tanh(4 * (data[0] - 0.5)) + 1)
            pts[0, :] = (p1_ * data[0] + p2_ * (1 - data[0]))
            pts[-1, :] = (p4_ * data[-1] + p3_ * (1 - data[-1]))
            pts[1:-1, :] += diff
            #ctr_pts = numpy.vstack((p1_, pts, p3_))
            ctr_pts = pts
        elif c1:
            diff = data[1:].reshape(-1, 3)
            t = numpy.linspace(0, 1, diff.shape[0] + 1)
            pts = ctlpts_func(t)
            pts[0, :] = (p1_ * 0.5 * (numpy.tanh(4 * (data[0] - 0.5)) + 1) +
                         pts[0, :] * (1 - 0.5 * (numpy.tanh(4 * (data[0] - 0.5)) + 1)))
            pts[1:, :] += diff
            #ctr_pts = numpy.vstack((p1_, pts))
            ctr_pts = pts
        elif c2:
            diff = data[:-1].reshape(-1, 3)
            t = numpy.linspace(0, 1, diff.shape[0] + 1)
            pts = ctlpts_func(t)
            pts[-1, :] = (p4_ * 0.5 * (numpy.tanh(4 * (data[-1] - 0.5)) + 1) +
                          pts[-1, :] * (1 - 0.5 * (numpy.tanh(4 * (data[-1] - 0.5)) + 1)))
            pts[:-1, :] += diff
            #ctr_pts = numpy.vstack((pts, p3_))
            ctr_pts = pts
        else:
            diff = data.reshape(-1, 3)
            t = numpy.linspace(0, 1, diff.shape[0])
            pts = ctlpts_func(t)
            ctr_pts = pts + diff
        """
        if c1 and c2:
            diff = data[2:-2].reshape(-1, 3)
            t = numpy.linspace(0.25, 0.75, diff.shape[0])
            pts = ctlpts_func(t)
            direction_start = (p2_ - p1_)
            direction_end = (p3_ - p4_)
            direction_start = direction_start / numpy.linalg.norm(direction_start)
            direction_end = direction_end / numpy.linalg.norm(direction_end)
            tmp_0 = 0.5 + 0.5 * 0.5 * (numpy.tanh(4 * (data[0] - 0.5)) + 1)
            tmp_1 = 0.5 + 0.5 * 0.5 * (numpy.tanh(4 * (data[-1] - 0.5)) + 1)
            tmp_0 = 0.5 + 0.5*abs(data[0])
            tmp_1 = 0.5 + 0.5*abs(data[-1])
            ctr_pt_start = p1_ * (1 - tmp_0) + tmp_0 * p2_
            ctr_pt_end = p3_ * (1 - tmp_1) + tmp_1 * p4_
            pts += diff
            tmp_2 = 0.5 + 0.5*0.5 * (numpy.tanh(4 * (data[1] - 0.5)) + 1)
            tmp_3 = 0.5 + 0.5*0.5 * (numpy.tanh(4 * (data[-2] - 0.5)) + 1)
            tmp_2 = abs(data[1])
            tmp_3 = abs(data[-2])
            #ctr_pt_start_vec = ctr_pt_start * (1 - tmp_2) + tmp_2 * p2_
            #ctr_pt_end_vec = ctr_pt_end * (1 - tmp_3) + tmp_3 * p4_
            #ctr_pts = numpy.vstack((ctr_pt_start, ctr_pt_start_vec, pts, ctr_pt_end_vec, ctr_pt_end))
            scale_1 = np.linalg.norm(p2_ - p1_)
            scale_2 = np.linalg.norm(p4_ - p3_)
            weights = [scale_1*tmp_2, scale_2*tmp_3]
            tangents = [direction_start, direction_end]
            ctr_pts = numpy.vstack((ctr_pt_start, ctr_pt_start + weights[0]*tangents[0], pts, ctr_pt_end - weights[1]*tangents[1], ctr_pt_end))
        elif c1:
            diff = data[2:].reshape(-1, 3)
            t = numpy.linspace(0.25, 0.75, diff.shape[0])
            pts = ctlpts_func(t)
            direction_start = (p2_ - p1_)
            direction_start = direction_start / numpy.linalg.norm(direction_start)
            tmp_0 = 0.5 + 0.5 * 0.5 * (numpy.tanh(4 * (data[0] - 0.5)) + 1)
            ctr_pt_start = p1_ * (1 - tmp_0) + tmp_0 * p2_
            pts += diff
            tmp_2 = 0.5 + 0.5*0.5 * (numpy.tanh(4 * (data[1] - 0.5)) + 1)
            pts += diff
            scale_1 = np.linalg.norm(p2_ - p1_)
            weights = [scale_1*tmp_2, None]
            tangents = [direction_start, None]
            ctr_pts = numpy.vstack((ctr_pt_start, ctr_pt_start + weights[0]*tangents[0], pts, p4_))
            weights = weights
            tangents = tangents
        elif c2:
            diff = data[:-2].reshape(-1, 3)
            t = numpy.linspace(0.25, 0.75, diff.shape[0])
            pts = ctlpts_func(t)
            direction_end = (p3_ - p4_)
            direction_end = direction_end / numpy.linalg.norm(direction_end)
            tmp_1 = 0.5 + 0.5 * 0.5 * (numpy.tanh(4 * (data[-1] - 0.5)) + 1)
            ctr_pt_end = p3_ * (1 - tmp_1) + tmp_1 * p4_
            pts += diff
            tmp_3 = 0.5 + 0.5*0.5 * (numpy.tanh(4 * (data[-2] - 0.5)) + 1)
            pts += diff
            scale_2 = np.linalg.norm(p4_ - p3_)
            weights = [None, scale_2*tmp_3]
            tangents = [None, direction_end]
            ctr_pts = numpy.vstack((p2_, pts, ctr_pt_end - weights[1]*tangents[1], ctr_pt_end))
            weights = weights
            tangents = tangents
        else:
            diff = data.reshape(-1, 3)
            t = numpy.linspace(0, 1, diff.shape[0]+2)[1:-1]
            pts = ctlpts_func(t)
            pts = pts + diff
            ctr_pts = numpy.vstack((p2_, pts, p4_))
            weights = None
            tangents = None
        #ctr_pts = ctr_pts.tolist()
        curve = Spline(ctr_pts, weights=weights, tangents=tangents)
        return curve
    return create_spline


def bezier_factory(p1, p2, p3, p4, clamp_first=True, clamp_second=True, ctlpts_func=None):
    def create_bezier(data, p1_=p1, p2_=p2, p3_=p3, p4_=p4, c1=clamp_first, c2=clamp_second, ctlpts_func=ctlpts_func):
        if c1 and c2:
            diff = data[2:-2].reshape(-1, 3)
            t = numpy.linspace(0.25, 0.75, diff.shape[0])
            pts = ctlpts_func(t)
            tmp_0 = 0.5 * (numpy.tanh(4 * (data[0] - 0.5)) + 1)
            tmp_1 = 0.5 * (numpy.tanh(4 * (data[-1] - 0.5)) + 1)
            ctr_pt_start = p1_ + (0.5 + 0.5 * tmp_0) * (p2_ - p1_)
            ctr_pt_end = p3_ + (0.5 + 0.5 * tmp_1) * (p4_ - p3_)
            pts += diff
            tmp_2 = 0.5 * (numpy.tanh(4 * (data[1] - 0.5)) + 1)
            tmp_3 = 0.5 * (numpy.tanh(4 * (data[-2] - 0.5)) + 1)
            ctr_pt_start_vec = ctr_pt_start + (0.1 + max(0.0, data[1])) * (p2_ - ctr_pt_start)
            ctr_pt_end_vec = ctr_pt_end + (0.1 + max(0.0, data[-2])) * (p4_ - ctr_pt_end)
            ctr_pts = numpy.vstack((ctr_pt_start, ctr_pt_start_vec, pts, ctr_pt_end_vec, ctr_pt_end))
        elif c1:
            diff = data[2:].reshape(-1, 3)
            t = numpy.linspace(0, 1, diff.shape[0] + 2)[1:-1]
            pts = ctlpts_func(t)
            ctr_pt_start = p1_ + data[0] * (p2_ - p1_)
            ctr_pt_start_vec = ctr_pt_start + data[1] * ((p2_ - p1_) / numpy.linalg.norm(p2_ - p1_))
            pts += diff
            ctr_pts = numpy.vstack((ctr_pt_start, ctr_pt_start_vec, pts, p4_))
        elif c2:
            diff = data[:-2].reshape(-1, 3)
            t = numpy.linspace(0, 1, diff.shape[0] + 2)[1:-1]
            pts = ctlpts_func(t)
            ctr_pt_end = p4_ + data[-1] * (p3_ - p4_)
            ctr_pt_end_vec = ctr_pt_end + data[-2] * (p4_ - p3_) / numpy.linalg.norm(p4_ - p3_)
            pts += diff
            ctr_pts = numpy.vstack((p2_, pts, ctr_pt_end_vec, ctr_pt_end))
        else:
            diff = data.reshape(-1, 3)
            t = numpy.linspace(0, 1, diff.shape[0]+2)[1:-1]
            pts = ctlpts_func(t)
            pts = pts + diff
            ctr_pts = numpy.vstack((p2_, pts, p4_))
        ctr_pts = ctr_pts.tolist()
        curve = BSpline.Curve()
        curve.degree = len(ctr_pts) - 1
        curve.ctrlpts = ctr_pts
        curve.knotvector = utilities.generate_knot_vector(curve.degree, len(curve.ctrlpts))
        return curve
    return create_bezier


def cost_bounds(bounds, pts, func=lambda x:max(0,x), boundary_func=None):
    bounds_score = 0.0
    if boundary_func is None:
        for i in range(pts.shape[0]):
            bounds_score += func(bounds[0, 0] - pts[i, 0])
            bounds_score += func(pts[i, 0] - bounds[0, 1])
            bounds_score += func(bounds[1, 0] - pts[i, 1])
            bounds_score += func(pts[i, 1] - bounds[1, 1])
            bounds_score += func(bounds[2, 0] - pts[i, 2])
            bounds_score += func(pts[i, 2] - bounds[2, 1])
    else:
        #print("Using boundary function")
        #for i in range(pts.shape[0]):
        #    bounds_score += func(boundary_func([pts[i, 0], pts[i, 1], pts[i, 2], bounds_k]))
        bounds_score = boundary_func(pts).flatten()
        bounds_score = bounds_score[bounds_score > 0]
        if len(bounds_score) > 0:
            bounds_score = numpy.max(bounds_score)
        else:
            bounds_score = 0.0
    return bounds_score


def collision_test(vessels, pts, radius, clearance):
    # this is not a perfect function and still needs to be optimized for accuracy in the line
    # segment distance calculation
    lines = numpy.zeros((pts.shape[0]-1, 6))
    lines[:, 0:3] = pts[:-1, :]
    lines[:, 3:6] = pts[1:, :]
    #line_centers = (lines[:, 0:3] + lines[:, 3:6]) / 2
    #vessel_centers = (vessels[:, 0:3] + vessels[:, 3:6]) / 2
    #vessel_lengths = numpy.linalg.norm(vessels[:, 3:6] - vessels[:, 0:3], axis=1).reshape(1, -1)
    #line_lengths = numpy.linalg.norm(lines[:, 3:6] - lines[:, 0:3], axis=1).reshape(-1, 1)
    #dists = numpy.linalg.norm(line_centers[:, numpy.newaxis, :] - vessel_centers[numpy.newaxis, :, :], axis=2)
    #dists = dists - (vessel_lengths + line_lengths) / 2
    #dists = numpy.min(dists, axis=1)
    dists = numpy.min(minimum_segment_distance(vessels, lines), axis=0)
    return numpy.sum(dists[dists < 2.0*radius + clearance])


def cost_new(data, curve_generator=None, r=None, collision_vessels=None, clearance=None, bounds=None, sample_size=None,
             boundary_func=None, length_threshold=None, length_bounds=None, start=None, end=None, angle_threshold=70,
             clamp_first=True, clamp_second=True):
    #start = perf_counter()
    curve = curve_generator(data)
    curve.sample_size = sample_size
    curve.evaluate()
    pts = numpy.array(curve.evalpts)
    #end = perf_counter()
    #print("Time to evaluate curve: ", end-start)
    #start = perf_counter()
    if isinstance(collision_vessels, type(None)):
        collision_cost = 0.0
    else:
        collision_cost = collision_test(collision_vessels, pts, r, clearance)
    #end = perf_counter()
    #print("Time to evaluate collisions: ", end-start)
    #start = perf_counter()

    if not isinstance(bounds, type(None)):
        bounds_cost = cost_bounds(bounds, pts)
    else:
        bounds_cost = 0.0
    if not isinstance(boundary_func, type(None)):
        bounds_cost = cost_bounds(bounds, pts, boundary_func=boundary_func)
    else:
        #print("no bounds")
        bounds_cost = 0.0
    #bounds_cost = 0.0
    #end = perf_counter()
    #print("Time to evaluate bounds: ", end-start)
    #start = perf_counter()
    u = numpy.linspace(0, 1, pts.shape[0])[1:-1]
    radius_of_curvature = []
    #derivatives_1 = []
    #derivatives_2 = []
    derivatives_1 = curve.derivatives(1).T
    first = derivatives_1[:-1, :]
    last = derivatives_1[1:, :]
    flips = np.linalg.norm(first*last, axis=1)
    flips = np.abs(np.sum(flips[flips < 0]))
    #derivatives_2 = curve.derivatives(2)[:, 1:-1]
    #derivatives_3 = curve.derivatives(3)[:, 1:-1]
    #for i in range(u.shape[0]):
    #    derivatives = curve.derivatives(u[i], 2)
    #    #derivatives_1.append(derivatives[1])
    #    #derivatives_2.append(derivatives[2])
    #    radius_of_curvature.append(1/(numpy.linalg.norm(numpy.cross(derivatives[1], derivatives[2]))/numpy.linalg.norm(derivatives[1])**3))
    #derivatives_1 = numpy.array(derivatives_1)
    #derivatives_2 = numpy.array(derivatives_2)
    #curvature = numpy.linalg.norm(numpy.cross(derivatives_1, derivatives_2, axis=0), axis=0)/numpy.linalg.norm(derivatives_1, axis=0)**3

    #torsion = numpy.linalg.norm(numpy.cross(derivatives_1, derivatives_2, axis=0) * derivatives_3, axis=0)/numpy.linalg.norm(numpy.cross(derivatives_1, derivatives_2, axis=0), axis=0)**2
    #radius_of_curvature.append(1/curvature)
    roc= 1/curve.curvature(4*sample_size)
    #radius_of_curvature = numpy.array(radius_of_curvature)
    if len(roc[roc < 2 * r]) == 0:
        curvature_cost = 0.0
    else:
        curvature_cost = (2*r - numpy.mean(roc[roc < 2 * r]))/(2*r)
    if clamp_first:
        vector_start = pts[0, :] - start
        vector_1 = pts[1, :] - pts[0, :]
        vector_start = vector_start / numpy.linalg.norm(vector_start)
        vector_1 = vector_1 / numpy.linalg.norm(vector_1)
        flip_1 = max(0.0, (numpy.arccos(numpy.dot(vector_start, vector_1)) * (180 / numpy.pi)) - angle_threshold)
        curvature_cost += np.tanh(flip_1)*0.1
    if clamp_second:
        vector_end = pts[-1, :] - end
        vector_2 = pts[-2, :] - pts[-1, :]
        vector_end = vector_end / numpy.linalg.norm(vector_end)
        vector_2 = vector_2 / numpy.linalg.norm(vector_2)
        flip_2 = max(0.0, (numpy.arccos(numpy.dot(vector_end, vector_2))*(180/numpy.pi)) - angle_threshold)
        curvature_cost += np.tanh(flip_2)*0.1
    #curvature_cost += flips
    #vectors = get_all_vectors(pts)
    #angles = get_all_angles(vectors)
    #angle_cost = cost_angles(angles-angle_threshold)
    #angle_cost = cost_angles(pts, angle_threshold)
    #end = perf_counter()
    #print("Time to evaluate curvature: ", end-start)
    #start = perf_counter()
    spline_length = numpy.sum(numpy.linalg.norm(numpy.diff(pts, axis=0), axis=1))
    if clamp_first:
        spline_length += numpy.linalg.norm(pts[0, :] - start)
    if clamp_second:
        spline_length += numpy.linalg.norm(pts[-1, :] - end)
    #end = perf_counter()
    #print("Time to evaluate length: ", end-start)
    length_cost = max(0.0, spline_length - length_bounds[1]*length_threshold)
    lower_length_cost = max(0.0, length_bounds[0]*length_threshold - spline_length)
    #print("Collision Cost: {}\tBounds Cost: {}\tCurvature Cost: {}\tLength Cost: {}".format(collision_cost,bounds_cost, curvature_cost, length_cost))
    pareto = np.tanh((np.tanh(max(0.0, abs(collision_cost))) + np.tanh(max(0.0, bounds_cost)) + np.tanh(max(0.0, curvature_cost)))) # + np.tanh(length_cost)))
    pareto_1 = max(0.0, abs(collision_cost))
    pareto_2 = max(0.0, bounds_cost)
    pareto_3 = max(0.0, curvature_cost)
    pareto_4 = np.tanh(length_cost/length_threshold)
    return (pareto +
            np.heaviside(pareto_1, 0)*2 +
            np.heaviside(pareto_2, 0)*4 +
            np.heaviside(pareto_3, 0) +
            np.heaviside(-pareto, 1) * np.heaviside(pareto_4, 0)*(0.5+0.5*pareto_4) +
            np.heaviside(-pareto, 1) * np.heaviside(-pareto_4, 1) * (0.5*np.tanh(lower_length_cost/length_threshold)))


class Connection(object):
    def __init__(self):
        self.proximal_0 = None
        self.distal_0 = None
        self.proximal_1 = None
        self.distal_1 = None
        self.radius_0 = None
        self.radius_1 = None
        self.solver = None
        self.bounds = None
        self.boundary_func = None
        self.ctrlpt_function = None
        self.curve_function = None
        self.physical_clearance = 0.0
        self.collision_vessels = None
        self.plotting_vessels = None
        self.clamp_first = True
        self.clamp_second = True
        self.history = []

    def set_first_vessel(self, proximal, distal, radius):
        self.proximal_0 = proximal
        self.distal_0 = distal
        self.radius_0 = radius

    def set_second_vessel(self, proximal, distal, radius):
        self.proximal_1 = proximal
        self.distal_1 = distal
        self.radius_1 = radius

    def set_solver(self, method='basinhopping'):
        if method == 'basinhopping':
            self.solver = optimize.basinhopping
        elif method == 'shgo':
            self.solver = optimize.shgo
        elif method == 'differential_evolution':
            self.solver = optimize.differential_evolution
        else:
            raise NotImplementedError("Solver method not implemented")

    def set_bounds(self, bounds, boundary_func=None):
        self.bounds = bounds
        self.boundary_func = boundary_func

    def set_ctrl_function(self, *args):
        """
        Set the function that seeds the control points for the initial
        guess of the optimization. The function should accept a vector
        of values between 0 and 1 and return a set of control points
        for the Bezier curve.
        """
        if len(args) == 1:
            self.ctrlpt_function = args[0]
        else:
            path_pts = numpy.vstack((self.distal_0, self.distal_1))
            k = 1
            tck = splprep(path_pts.T, s=0, k=k)
            self.ctrlpt_function = lambda t_: numpy.array(splev(t_, tck[0])).T

    def set_physical_clearance(self, clearance):
        self.physical_clearance = clearance

    def set_collision_vessels(self, collision_vessels):
        self.collision_vessels = collision_vessels

    def solve(self, *args, sample_size=20, **kwargs):
        self.history_x = []
        self.history_f = []
        def callback(x, f, accept):
            self.history_x.append(x)
            self.history_f.append(f)
            if f == 0:
                return True
        length_bounds = kwargs.get('length_bounds', [0.5, 1.05])
        angle_threshold = kwargs.get('angle_threshold', 70.0)
        if len(args) == 0:
            number_free_points = 1
        else:
            number_free_points = args[0]
        #if isinstance(self.ctrlpt_function, type(None)):
        #    path_pts = numpy.vstack((self.distal_0, self.distal_1))
        #    k = 1
        #    tck = splprep(path_pts.T, s=0, k=k)
        #    self.ctrlpt_function = lambda t_: numpy.array(splev(t_, tck[0])).T
        u_fine = numpy.linspace(0, 1, 25)
        fine_pts = self.ctrlpt_function(u_fine)
        length_threshold = numpy.sum(numpy.linalg.norm(numpy.diff(fine_pts, axis=0), axis=1))
        init_spline_length = deepcopy(length_threshold)
        if self.clamp_first:
            length_threshold += numpy.linalg.norm(self.distal_0-self.proximal_0)
        if self.clamp_second:
            length_threshold += numpy.linalg.norm(self.distal_1-self.proximal_1)
        self.curve_function = spline_factory(self.proximal_0, self.distal_0, self.proximal_1, self.distal_1,
                                             clamp_first=self.clamp_first, clamp_second=self.clamp_second,
                                             ctlpts_func=self.ctrlpt_function)
        cost_function = lambda data: cost_new(data, curve_generator=self.curve_function, r=self.radius_0,
                                              collision_vessels=self.collision_vessels, clearance=self.physical_clearance,
                                              bounds=self.bounds, sample_size=sample_size, boundary_func=self.boundary_func,
                                              length_threshold=length_threshold, length_bounds=length_bounds, start=self.proximal_0,
                                              end=self.proximal_1, angle_threshold=angle_threshold, clamp_first=self.clamp_first,
                                              clamp_second=self.clamp_second)
        if self.clamp_first and self.clamp_second:
            xopt = numpy.zeros(4+number_free_points*3)
            xopt[0] = 0.5
            xopt[1] = 1.0
            xopt[-1] = 0.5
            xopt[-2] = 1.0
        elif self.clamp_first or self.clamp_second:
            xopt = numpy.zeros(2+number_free_points*3)
            if self.clamp_first:
                xopt[0] = 0.5
                xopt[1] = 1.0
            else:
                xopt[-1] = 0.5
                xopt[-2] = 1.0
        else:
            xopt = numpy.zeros(number_free_points*3)
        shortest_length = numpy.linalg.norm(self.distal_0-self.distal_1)
        #lb = numpy.ones(xopt.shape[0]) * (-shortest_length)
        #ub = numpy.ones(xopt.shape[0]) * (shortest_length)
        lb = [None] * xopt.shape[0]
        ub = [None] * xopt.shape[0]
        if self.clamp_first:
            lb[0] = 0
            lb[1] = 0
            ub[0] = 1
        if self.clamp_second:
            lb[-1] = 0
            lb[-2] = 0
            ub[-1] = 1
        solver_bounds = []
        for b in range(len(lb)):
            solver_bounds.append([lb[b], ub[b]])
        res = self.solver(cost_function, xopt, niter=500, callback=callback, stepsize=init_spline_length/(number_free_points+1),
                          minimizer_kwargs={'bounds': solver_bounds, 'method': 'L-BFGS-B', 'tol': 1e-5})
        self.xopt = res.x
        self.best = res.fun
        self.curve = self.curve_function(self.xopt)
        self.curve.sample_size = sample_size
        self.curve.evaluate()
        #res.elapsed_time = perf_counter() - start
        return res

    def build_vessels(self, seperate, build_meshes=True):
        pts = numpy.array(self.curve.evalpts)
        #pts = self.curve.evalpts
        if seperate:
            sep = (pts.shape[0]-1)//2
            vessels_1 = numpy.zeros((sep,7))
            vessels_2 = numpy.zeros((pts.shape[0]-1-sep,7))
            vessels_1_pts = pts[:sep+1,:]
            vessels_2_pts = numpy.flip(pts[sep:,:],axis=0)
            vessels_1[:,0:3] = vessels_1_pts[:-1,:]
            vessels_1[:,3:6] = vessels_1_pts[1:,:]
            vessels_2[:,0:3] = vessels_2_pts[:-1,:]
            vessels_2[:,3:6] = vessels_2_pts[1:,:]
            vessels_1[:,6] = self.radius_0
            vessels_2[:,6] = self.radius_1
        else:
            vessels_1 = numpy.zeros((pts.shape[0]-1,7))
            vessels_1[:,0:3] = pts[:-1,:]
            vessels_1[:,3:6] = pts[1:,:]
            vessels_1[:,6] = self.radius_0
            pts_reverse = numpy.flip(pts,axis=0)
            vessels_2 = numpy.zeros((pts.shape[0]-1,7))
            vessels_2[:,0:3] = pts_reverse[:-1,:]
            vessels_2[:,3:6] = pts_reverse[1:,:]
            vessels_2[:,6] = self.radius_1
        self.vessels_1 = vessels_1
        self.vessels_2 = vessels_2
        self.seperate = seperate
        self.vessel_1_meshes = []
        if build_meshes:
            for i in range(vessels_1.shape[0]):
                center = (self.vessels_1[i, 0:3] + self.vessels_1[i, 3:6]) / 2
                direction = self.vessels_1[i, 3:6] - self.vessels_1[i, 0:3]
                length = numpy.linalg.norm(direction)
                direction = direction / length
                radius = self.vessels_1[i, 6]
                cylinder = pv.Cylinder(center=center, direction=direction, radius=radius, height=length)
                self.vessel_1_meshes.append(cylinder)
            self.vessel_2_meshes = []
            for i in range(self.vessels_2.shape[0]):
                center = (self.vessels_2[i, 0:3] + self.vessels_2[i, 3:6]) / 2
                direction = self.vessels_2[i, 3:6] - self.vessels_2[i, 0:3]
                length = numpy.linalg.norm(direction)
                direction = direction / length
                radius = self.vessels_2[i, 6]
                cylinder = pv.Cylinder(center=center, direction=direction, radius=radius, height=length)
                self.vessel_2_meshes.append(cylinder)
        return vessels_1, vessels_2

    def show(self):
        plotter = pv.Plotter()
        if self.seperate:
            for i in range(self.vessels_1.shape[0]):
                center = (self.vessels_1[i,0:3]+self.vessels_1[i,3:6])/2
                direction = self.vessels_1[i,3:6] - self.vessels_1[i,0:3]
                length = numpy.linalg.norm(direction)
                direction = direction/length
                radius = self.vessels_1[i,6]
                cylinder = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
                plotter.add_mesh(cylinder,color='red')
            for i in range(self.vessels_2.shape[0]):
                center = (self.vessels_2[i,0:3]+self.vessels_2[i,3:6])/2
                direction = self.vessels_2[i,3:6] - self.vessels_2[i,0:3]
                length = numpy.linalg.norm(direction)
                direction = direction/length
                radius = self.vessels_2[i,6]
                cylinder = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
                plotter.add_mesh(cylinder,color='blue')
        else:
            for i in range(self.vessels_1.shape[0]):
                center = (self.vessels_1[i,0:3]+self.vessels_1[i,3:6])/2
                direction = self.vessels_1[i,3:6] - self.vessels_1[i,0:3]
                length = numpy.linalg.norm(direction)
                direction = direction/length
                radius = self.vessels_1[i,6]
                cylinder = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
                plotter.add_mesh(cylinder,color='red')
        if self.plotting_vessels is not None:
            for i in range(self.plotting_vessels.shape[0]):
                center = (self.plotting_vessels[i,0:3]+self.plotting_vessels[i,3:6])/2
                direction = self.plotting_vessels[i,3:6] - self.plotting_vessels[i,0:3]
                length = numpy.linalg.norm(direction)
                direction = direction/length
                radius = self.plotting_vessels[i,6]
                cylinder = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
                plotter.add_mesh(cylinder,color='black')
        if np.all(self.proximal_0 != self.distal_0):
            center = (self.proximal_0+self.distal_0)/2
            direction = (self.distal_0-self.proximal_0)
            length = numpy.linalg.norm(direction)
            direction = direction/length
            radius = self.radius_0
            cylinder = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
            plotter.add_mesh(cylinder,color='yellow')
        if np.all(self.proximal_1 != self.distal_1):
            center = (self.proximal_1+self.distal_1)/2
            direction = (self.distal_1-self.proximal_1)
            length = numpy.linalg.norm(direction)
            direction = direction/length
            radius = self.radius_1
            cylinder = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
            plotter.add_mesh(cylinder,color='yellow')
        return plotter

    def show_history(self):
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        plt.subplots_adjust(left=0.1, bottom=0.25)
        line, = ax.plot([], [], [], 'b-', label='Connection History')
        spline = self.curve_function(self.history_x[0])
        spline.sample_size = 20
        spline.evaluate()
        his = spline.evalpts
        for x in self.history_x[1:]:
            spline = self.curve_function(x)
            spline.sample_size = 20
            spline.evaluate()
            his = np.vstack((his, spline.evalpts))
        x_min = np.min(his[:, 0])
        x_max = np.max(his[:, 0])
        y_min = np.min(his[:, 1])
        y_max = np.max(his[:, 1])
        z_min = np.min(his[:, 2])
        z_max = np.max(his[:, 2])

        def update_plot(idx):
            spline = self.curve_function(self.history_x[idx])
            spline.sample_size = 20
            spline.evaluate()
            upstream_1 = numpy.vstack([self.proximal_0, spline.evalpts[0, :]])
            upstream_2 = numpy.vstack([spline.evalpts[-1, :], self.proximal_1])
            ax.clear()
            ax.plot(spline.ctrlpts[:, 0], spline.ctrlpts[:, 1], spline.ctrlpts[:, 2], 'yo--', label='Control Points')
            ax.plot(spline.evalpts[:, 0], spline.evalpts[:, 1], spline.evalpts[:, 2], 'g-', label='Connection')
            ax.plot(upstream_1[:, 0], upstream_1[:, 1], upstream_1[:, 2], 'r-', label='Tree 1')
            ax.plot(upstream_2[:, 0], upstream_2[:, 1], upstream_2[:, 2], 'b-', label='Tree 2')
            ax.set_xlim([x_min, x_max])
            ax.set_ylim([y_min, y_max])
            ax.set_zlim([z_min, z_max])
            ax.legend()
            ax.set_xlabel('X')
            ax.set_ylabel('Y')
            ax.set_zlabel('Z')
            if (self.history_f[idx] // 7) == 1:
                collision = True
                bounds = True
                curvature = True
            elif (self.history_f[idx] // 6) == 1:
                collision = True
                bounds = True
                curvature = False
            elif (self.history_f[idx] // 5) == 1:
                collision = False
                bounds = True
                curvature = True
            elif (self.history_f[idx] // 4) == 1:
                collision = False
                bounds = True
                curvature = False
            elif (self.history_f[idx] // 3) == 1:
                collision = True
                bounds = False
                curvature = True
            elif (self.history_f[idx] // 2) == 1:
                collision = True
                bounds = False
                curvature = False
            elif (self.history_f[idx] // 1) == 1:
                collision = False
                bounds = False
                curvature = True
            elif self.history_f[idx] < 1:
                collision = False
                bounds = False
                curvature = False
            else:
                raise ValueError("Invalid cost function")
            title = 'Connection History at Iteration {}, Cost: {} \n Collision: {} ,Boundary: {}, Curvature: {}'.format(idx, self.history_f[idx], collision, bounds, curvature)
            ax.set_title(title)
            fig.canvas.draw_idle()

        ax_slider = fig.add_axes([0.1, 0.1, 0.8, 0.03], facecolor='lightgoldenrodyellow')
        slider = Slider(ax_slider, 'Control Set', 0, len(self.history_x) - 1, valinit=0, valfmt='%0.0f')
        def update(val):
            idx = int(slider.val)
            update_plot(idx)

        slider.on_changed(update)
        update_plot(0)
        plt.show()

class SimpleConnection(object):
    def __init__(self, forest, network_id, tree_0, tree_1, idx, jdx, ctrl_function=None, clamp_first=True,
                 clamp_second=True, point_0=None, point_1=None):
        self.forest = forest
        conn = Connection()
        vessel_0 = forest.networks[network_id][tree_0].data[idx, :]
        vessel_1 = forest.networks[network_id][tree_1].data[jdx, :]
        if not isinstance(point_0, type(None)):
            conn.set_first_vessel(point_0[0:3], point_0[3:6], vessel_0[21])
        else:
            conn.set_first_vessel(vessel_0[0:3], vessel_0[3:6], vessel_0[21])
        if not isinstance(point_1, type(None)):
            conn.set_second_vessel(point_1[0:3], point_1[0:3], vessel_1[21])
        else:
            conn.set_second_vessel(vessel_1[0:3], vessel_1[3:6], vessel_1[21])
        collision_vessels = []
        tree_0_idx = numpy.arange(forest.networks[network_id][tree_0].data.shape[0], dtype=int).tolist()
        if not numpy.isnan(vessel_0[17]):
            parent = int(vessel_0[17])
            daughter_0 = int(forest.networks[network_id][tree_0].data[parent, 15])
            daughter_1 = int(forest.networks[network_id][tree_0].data[parent, 16])
            tree_0_idx.remove(parent)
            tree_0_idx.remove(daughter_0)
            tree_0_idx.remove(daughter_1)
        if not numpy.isnan(vessel_0[15]):
            daughter_2 = int(vessel_1[15])
            tree_0_idx.remove(daughter_2)
        if not numpy.isnan(vessel_0[16]):
            daughter_3 = int(vessel_1[16])
            tree_0_idx.remove(daughter_3)
        tree_0_idx = numpy.array(tree_0_idx).astype(int)
        tmp = numpy.zeros((tree_0_idx.shape[0], 7))
        tmp[:, 0:3] = forest.networks[network_id][tree_0].data[tree_0_idx, 0:3]
        tmp[:, 3:6] = forest.networks[network_id][tree_0].data[tree_0_idx, 3:6]
        tmp[:, 6] = forest.networks[network_id][tree_0].data[tree_0_idx, 21]
        collision_vessels.append(tmp)
        tree_1_idx = numpy.arange(forest.networks[network_id][tree_1].data.shape[0], dtype=int).tolist()
        if not numpy.isnan(vessel_1[17]):
            parent = int(vessel_1[17])
            daughter_0 = int(forest.networks[network_id][tree_1].data[parent, 15])
            daughter_1 = int(forest.networks[network_id][tree_1].data[parent, 16])
            tree_1_idx.remove(parent)
            tree_1_idx.remove(daughter_0)
            tree_1_idx.remove(daughter_1)
        if not numpy.isnan(vessel_1[15]):
            daughter_2 = int(vessel_1[15])
            tree_1_idx.remove(daughter_2)
        if not numpy.isnan(vessel_1[16]):
            daughter_3 = int(vessel_1[16])
            tree_1_idx.remove(daughter_3)
        tree_1_idx = numpy.array(tree_1_idx).astype(int)
        tmp = numpy.zeros((tree_1_idx.shape[0], 7))
        tmp[:, 0:3] = forest.networks[network_id][tree_1].data[tree_1_idx, 0:3]
        tmp[:, 3:6] = forest.networks[network_id][tree_1].data[tree_1_idx, 3:6]
        tmp[:, 6] = forest.networks[network_id][tree_1].data[tree_1_idx, 21]
        collision_vessels.append(tmp)
        for i in range(forest.n_networks):
            for j in range(forest.n_trees_per_network[i]):
                if i == network_id and (j == tree_0 or j == tree_1):
                    continue
                tmp = numpy.zeros((forest.networks[i][j].data.shape[0], 7))
                tmp[:, 0:3] = forest.networks[i][j].data[:, 0:3]
                tmp[:, 3:6] = forest.networks[i][j].data[:, 3:6]
                tmp[:, 6] = forest.networks[i][j].data[:, 21]
                collision_vessels.append(tmp)
        collision_vessels = numpy.vstack(collision_vessels)
        if collision_vessels.shape[0] > 0:
            conn.set_collision_vessels(collision_vessels)
        conn.set_solver()
        conn.set_physical_clearance(self.forest.physical_clearance)
        bounds = numpy.zeros((3, 2))
        bounds[:, 0] = numpy.min(forest.domain.points, axis=0).T
        bounds[:, 1] = numpy.max(forest.domain.points, axis=0).T
        bounds_func = forest.domain.__call__
        conn.set_bounds(bounds, boundary_func=bounds_func)
        if not isinstance(ctrl_function, type(None)):
            conn.set_ctrl_function(ctrl_function)
        conn.clamp_first = clamp_first
        conn.clamp_second = clamp_second
        self.connection = conn

    def solve(self, *args, **kwargs):
        return self.connection.solve(*args, **kwargs)

    def show(self):
        plotter = self.forest.show(plot_domain=True, return_plotter=True)
        self.connection.build_vessels(True)
        for mesh in self.connection.vessel_1_meshes:
            plotter.add_mesh(mesh, color='red')
        for mesh in self.connection.vessel_2_meshes:
            plotter.add_mesh(mesh, color='blue')
        return plotter

    def show_init(self):
        plotter = self.forest.show(plot_domain=True, return_plotter=True)
        if not isinstance(self.connection.xopt, type(None)):
            xopt = self.connection.xopt.copy()
        self.connection.xopt = np.zeros_like(xopt)
        self.connection.xopt[0] = 0.5
        self.connection.xopt[1] = 1.0
        self.connection.xopt[-1] = 0.5
        self.connection.xopt[-2] = 1.0
        original_curve = deepcopy(self.connection.curve)
        original_sample_size = deepcopy(original_curve.sample_size)
        self.connection.curve = self.connection.curve_function(self.connection.xopt)
        self.connection.curve.sample_size = original_sample_size
        self.connection.curve.evaluate()
        self.connection.build_vessels(True)
        for mesh in self.connection.vessel_1_meshes:
            plotter.add_mesh(mesh, color='red')
        for mesh in self.connection.vessel_2_meshes:
            plotter.add_mesh(mesh, color='blue')
        self.connection.xopt = xopt
        self.connection.curve = original_curve
        self.connection.curve.sample_size = original_sample_size
        self.connection.curve.evaluate()
        self.connection.build_vessels(True)
        return plotter

class TreeConnections(object):
    def __init__(self, forest, network_id):
        self.forest = forest
        self.network_id = network_id
        self.assignments, self.connections = assign_network(forest, network_id)
        self.vessel_connections = []
        self.connection_collisions = []
        self.connecting_vessels = []
        self.network = []
        for i in range(forest.n_trees_per_network[network_id]):
            self.network.append(deepcopy(forest.networks[network_id][i].data))

    def solve(self, *args, **kwargs):
        self.vessel_connections = []
        self.connection_collisions = []
        self.connecting_vessels = []
        tree_0 = 0
        tree_1 = 1
        tree_connections = []
        midpoints = []
        for j in trange(len(self.connections[0]), desc=f"Tree {tree_0} to Tree {tree_1}", leave=False):
            self.connecting_vessels.append([])
            self.connecting_vessels.append([])
            conn = SimpleConnection(self.forest, self.network_id, tree_0, tree_1,
                                    self.assignments[tree_0][j], self.assignments[tree_1][j],
                                    ctrl_function=self.connections[0][j],
                                    clamp_first=True, clamp_second=True)
            if len(self.connection_collisions) > 0:
                collisions = numpy.vstack(self.connection_collisions)
                if not isinstance(conn.connection.collision_vessels, type(None)):
                    collisions = numpy.vstack((collisions, conn.connection.collision_vessels))
                conn.connection.set_collision_vessels(collisions)
            conn.solve(*args, **kwargs)
            conn.connection.build_vessels(True, build_meshes=False)
            self.connection_collisions.append(conn.connection.vessels_1)
            self.connection_collisions.append(conn.connection.vessels_2)
            self.connecting_vessels[tree_0].append(conn.connection.vessels_1)
            self.connecting_vessels[tree_1].append(conn.connection.vessels_2)
            tree_connections.append(conn)
            midpoints.append(conn.connection.vessels_1[-1, 3:6])
        midpoints = numpy.array(midpoints)
        self.vessel_connections.append(tree_connections)
        if self.forest.n_trees_per_network[self.network_id] > 2:
            remaining_assignments, remaining_connections = assign_network_vector(self.forest, self.network_id, midpoints)
            self.assignments.extend(remaining_assignments)
            self.connections.extend(remaining_connections)
            for n in range(2, self.forest.n_trees_per_network[self.network_id]):
                tree_0 = 0
                tree_n = n
                tree_connections = []
                self.connecting_vessels.append([])
                for j in trange(len(remaining_connections[n-2]), desc=f"Tree {tree_0} to Tree {tree_n}", leave=False):
                    conn = SimpleConnection(self.forest, self.network_id, tree_n, tree_0,
                                            remaining_assignments[n-2][j], self.assignments[tree_0][j],
                                            ctrl_function=remaining_connections[n-2][j],
                                            clamp_first=True, clamp_second=False, point_1=midpoints[j])
                    #if len(self.connection_collisions) > 0:
                    #    collisions = numpy.vstack(self.connection_collisions)
                    #    if not isinstance(conn.connection.collision_vessels, type(None)):
                    #        collisions = numpy.vstack((collisions, conn.connection.collision_vessels))
                    #    conn.connection.set_collision_vessels(collisions)
                    collisions = []
                    for i in range(len(self.connecting_vessels)):
                        for c in range(len(self.connecting_vessels[i])):
                            if c == j:
                                collisions.append(self.connecting_vessels[i][c][:-1, :])
                            else:
                                collisions.append(self.connecting_vessels[i][c])
                    collisions = numpy.vstack(collisions)
                    #conn.connection.set_collision_vessels(collisions)
                    conn.solve(*args, **kwargs)
                    conn.connection.build_vessels(False, build_meshes=False)
                    self.connection_collisions.append(conn.connection.vessels_1)
                    self.connecting_vessels[tree_n].append(conn.connection.vessels_1)
                    tree_connections.append(conn)
                self.vessel_connections.append(tree_connections)
        """
        for i in range(len(self.connections)):
            tree_0 = 0
            tree_1 = i + 1
            if i == 0:
                # Account for the split bewteen "arterial/venous" of first connection
                self.connecting_vessels.append([])
                self.connecting_vessels.append([])
            else:
                self.connecting_vessels.append([])
            for j in trange(len(self.connections[0]), desc=f"Tree {tree_0} to Tree {tree_1}"):
                if i == 0:
                    conn = SimpleConnection(self.forest, self.network_id, tree_0, tree_1,
                                            self.assignments[tree_0][j], self.assignments[tree_1][j])
                else:
                    conn = SimpleConnection(self.forest, self.network_id, tree_0, tree_1,
                                            self.assignments[tree_0][j], self.assignments[tree_1][j],
                                            ctrl_function=self.connections[i][j], clamp_first=True,
                                            clamp_second=False)
                if len(self.connection_collisions) > 0:
                    collisions = numpy.vstack(self.connection_collisions)
                    if not isinstance(conn.connection.collision_vessels, type(None)):
                        collisions = numpy.vstack((collisions, conn.connection.collision_vessels))
                    conn.connection.set_collision_vessels(collisions)
                conn.solve(*args, **kwargs)
                if i == 0:
                    conn.connection.build_vessels(True, build_meshes=False)
                    self.connection_collisions.append(conn.connection.vessels_1)
                    self.connection_collisions.append(conn.connection.vessels_2)
                    self.connecting_vessels[tree_0].append(conn.connection.vessels_1)
                    self.connecting_vessels[tree_1].append(conn.connection.vessels_2)
                else:
                    conn.connection.build_vessels(False, build_meshes=False)
                    self.connection_collisions.append(conn.connection.vessels_2)
                    self.connecting_vessels[tree_1].append(conn.connection.vessels_2)
                self.vessel_connections.append(conn)
        """

    def show(self, **kwargs):
        colors = kwargs.get('colors', ['red', 'blue', 'green', 'yellow', 'purple',
                                       'orange', 'cyan', 'magenta', 'white', 'black'])
        plotter = pv.Plotter(**kwargs)
        count = 0
        for i in range(len(self.network)):
            for j in range(self.network[i].shape[0]):
                if j in self.assignments[i]:
                    conn_id = self.assignments[i].index(j)
                    center = (self.network[i][j, 0:3] + self.connecting_vessels[i][conn_id][0, 0:3]) / 2
                    length = numpy.linalg.norm(self.connecting_vessels[i][conn_id][0, 0:3] - self.network[i][j, 0:3])
                    direction = (self.connecting_vessels[i][conn_id][0, 0:3] - self.network[i][j, 0:3]) / length
                    radius = self.network[i][j, 21]
                    vessel = pv.Cylinder(center=center, direction=direction, radius=radius, height=length)
                else:
                    center = (self.network[i][j, 0:3] + self.network[i][j, 3:6]) / 2
                    direction = self.network[i].get('w_basis', j)
                    radius = self.network[i].get('radius', j)
                    length = self.network[i].get('length', j)
                    vessel = pv.Cylinder(center=center, direction=direction, radius=radius, height=length)
                plotter.add_mesh(vessel, color=colors[count % len(colors)], opacity=0.25)
            count += 1
        count = 0
        for conn in range(len(self.connecting_vessels)):
            for i in range(len(self.connecting_vessels[conn])):
                for j in range(self.connecting_vessels[conn][i].shape[0]):
                    center = (self.connecting_vessels[conn][i][j, 0:3] + self.connecting_vessels[conn][i][j, 3:6]) / 2
                    direction = self.connecting_vessels[conn][i][j, 3:6] - self.connecting_vessels[conn][i][j, 0:3]
                    length = numpy.linalg.norm(direction)
                    direction = direction / length
                    radius = self.connecting_vessels[conn][i][j, 6]
                    cylinder = pv.Cylinder(center=center, direction=direction, radius=radius, height=length)
                    plotter.add_mesh(cylinder, color=colors[count % len(colors)])
            count += 1
        plotter.add_mesh(self.forest.domain.boundary, color='grey', opacity=0.15)
        return plotter


class ForestConnections(object):
    def __init__(self, forest):
        self.forest = forest
        self.network_connections = []
        for i in range(forest.n_networks):
            self.network_connections.append(TreeConnections(forest, i))

    def solve(self, network_id, *args, **kwargs):
        for i in range(len(self.network_connections)):
            self.network_connections[i].solve(*args, **kwargs)

    def show(self, **kwargs):
        plotter = pv.Plotter(**kwargs)
        for i in range(len(self.network_connections)):
            self.network_connections[i].show(plotter=plotter)
        return plotter


# Copyright (c) Stanford University, The Regents of the University of
#               California, and others.
#
# All Rights Reserved.
#
# See Copyright-SimVascular.txt for additional details.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject
# to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
This module is used to create an input file for the SimVascular 1D solver (https://github.com/SimVascular/oneDSolver).

A 1D mesh is generated from the centerline geometry caculated from a closed polygonal surface.

A centerline consists of m cells, m=number of tract ids, the length of a cell/line is an approximation of a group.
In Cell Data, lines are listed from 0 to m. For each line, the first number is the number of points for this line
followed by 1st point to the last point.

"""
import os
import logging
import re
import numpy as np
from collections import OrderedDict

from scipy.signal import argrelextrema
import vtk
from vtk.util.numpy_support import vtk_to_numpy as v2n
from vtk import vtkIdList
from vtk import vtkPoints, vtkLine, vtkCellArray, vtkPolyData, vtkXMLPolyDataWriter

from .manage import get_logger_name
from .parameters import OutflowBoundaryConditionType, MaterialModel
from .utils import SurfaceFileFormats, read_polydata, write_polydata
from .io_1d import read_inflow_file, read_outlet_face_names, write_mesh, write_solver_file, \
    write_solver_nodes, write_solver_joints, write_solver_segments, write_solver_options, \
    write_solver_material, write_solver_output, write_solver_section_header, coronary_sv_to_oned, \
    read_variable_outflow_bcs
from .models import ZeroD

from .io_0d import write_0d_solver_file

import pdb


class Mesh(object):
    """
    The Mesh class is used to encapsulate 1D mesh calculations.
    """

    class CellDataFields(object):
        """ This class defines the required field cell data field names.
        """
        pass

    class PointDataFields(object):
        """ This class defines the required field point data field names.
        """
        AREA = "CenterlineSectionArea"
        CENT = "CenterlineId"
        PATH = "Path"
        BRANCH = "BranchId"
        BIFURCATION = "BifurcationId"
        NODEID = "GlobalNodeId"
        NORMAL = "CenterlineSectionNormal"

    class Open(object):
        """ This class wraps the 'open' class and adds a method to automatically
            write newlines.
        """

        def __init__(self, *args, **kwds):
            self.args = args
            self.kwds = kwds
            self.file_obj = open(*self.args, **self.kwds)

        def __enter__(self):
            self.file_obj = open(*self.args, **self.kwds)
            return self

        def __exit__(self, *args):
            self.file_obj.close()

        def close(self):
            self.file_obj.close()

        def writeln(self, string):
            self.file_obj.write(string + '\n')

        def write(self, string):
            self.file_obj.write(string)

    def __init__(self):
        self.centerlines = None
        self.logger = logging.getLogger(get_logger_name())

        # centerline
        self.centerline = None
        self.centerlines_outlet_face_names = None

        # caps
        self.inflow_data = None
        self.outlet_face_names = None
        self.outlet_face_names_index = None

        # boundary conditions
        self.bc_map = None
        self.bc_type = None

        # discretization
        self.seg_connectivity = None
        self.cell_data = None
        self.point_data = None
        self.seg_rear = None
        self.terminal = None
        self.junctions = None

        # discretization statistics
        self.num_seg = None
        self.num_bf_geo = None
        self.num_bf_branch = None
        self.num_elements = None

        # Some constansts for writing the solver file.
        self.solver_file_msg = "\n\n### DO NOT CHANGE THIS SECTION - generated automatically"
        self.space = " "
        self.endl = "\n"

    def generate(self, params, centerlines):
        """ Generate a mesh.
        """
        self.logger.info("Generate the 1D mesh ...")
        self.centerline = centerlines.geometry
        self.centerlines_outlet_face_names = centerlines.outlet_face_names

        # Set outlet face names.
        self.set_outlet_face_names(params)
        self.logger.info("Outlet face names: %s" % str(self.outlet_face_names))

        # check that centerline consist of one region
        if not self.check_centerline_region():
            return False

        # Check that centerline geometry has the required data fields.
        if not self.check_centerlines_data():
            return False

        # import module for adaptive meshing
        if params.seg_size_adaptive and params.seg_min_num > 1:
            try:
                import pwlf
            except ImportError:
                params.seg_size_adaptive = False
                self.logger.info('Could not import pwlf for adaptive meshing. Using uniform mesh instead')

        # mesh discretization
        self.discretize(params)
        self.logger.info("Number of centerline branches: %d" % len(self.cell_data['BranchId']))
        self.logger.info("Number of centerline bifurcations: %d" % self.num_bf_geo)
        self.logger.info("Number of svOneDSolver branches: %d" % self.num_seg)
        self.logger.info("Number of svOneDSolver joints: %d" % self.num_bf_branch)

        if not params.uniform_bc:
            self.set_variable_outflow_bcs(params)

        if params.inflow_input_file:
            read_inflow_file(self, params)

        if not params.uniform_material:
            msg = 'Non-uniform material not implemented'
            self.logger.error(msg)
            raise RuntimeError(msg)

        if params.model_order == 0:
            zerod = ZeroD(self, params)
            model = zerod.generate()

            # statistics
            self.num_elements = self.num_seg

            # write svZeroDSolver input file (.json)
            write_0d_solver_file(self, params, model)

        elif params.model_order == 1:
            # write discretization for visualization in ParaView
            write_mesh(self, params)

            # write svOneDSolver input file
            write_solver_file(self, params)

        return True

    def set_outlet_face_names(self, params):
        """ Set outlet face names.
        """
        if self.centerlines_outlet_face_names is not None:
            outlet_face_names = self.centerlines_outlet_face_names
        else:
            outlet_face_names = read_outlet_face_names(self, params)

        # Create a map between outlet face name and path ID.
        self.outlet_face_names_index = OrderedDict()
        self.outlet_face_names = []
        for i, face_name in enumerate(outlet_face_names):
            self.outlet_face_names.append(face_name)
            self.outlet_face_names_index[face_name] = i

    def set_variable_outflow_bcs(self, params):
        """ Read in data for variable flow boundary conditions.

        Parameters for outflow BCs (resistance or RCR) are read from a file.
        """
        self.logger.info("Set variable outflow boundary conditions ...")

        # convert legacy parameters
        if isinstance(params.outflow_bc_type, str):
            params.outflow_bc_type = [OutflowBoundaryConditionType.ONED_TO_SV[params.outflow_bc_type]]
            params.outflow_bc_file = os.path.dirname(params.outflow_bc_file)

        try:
            self.bc_map, self.bc_type = read_variable_outflow_bcs(params)
        except Exception as e:
            self.logger.error('Error while reading variable outflow boundary conditions')
            raise RuntimeError(str(e))

        if len(self.bc_map) != len(self.outlet_face_names):
            msg = "The number of BC values %d do not match the number of outlets %d." % (
                len(self.bc_map), len(self.outlet_face_names))
            raise RuntimeError(msg)

        for s, t in self.bc_type.items():
            self.logger.info('Surface ' + s + ' boundary condition type ' + t)

    def check_centerline_region(self):
        """
        Check if centerline is one connected region
        """
        con = vtk.vtkConnectivityFilter()
        con.SetInputData(self.centerline)
        con.SetExtractionModeToAllRegions()
        con.ColorRegionsOn()
        con.Update()
        n_region = con.GetNumberOfExtractedRegions()

        if n_region == 0:
            self.logger.error('Centerline is empty')
            raise RuntimeError('Centerline is empty')
        if n_region > 1:
            self.logger.error('Centerline consist of more than one region')
            raise RuntimeError('Centerline consist of more than one region')

        return True

    def check_centerlines_data(self):
        """ Check that the centerline data contains all of the required fields.
        """
        field_names = [v for k, v in self.CellDataFields.__dict__.items() if not k.startswith('__')]
        for field in field_names:
            if not self.centerline.GetCellData().GetArray(field):
                self.logger.error("Centerlines do not contain the '%s' data field." % field)
                return False

        field_names = [v for k, v in self.PointDataFields.__dict__.items() if not k.startswith('__')]
        for field in field_names:
            if not self.centerline.GetPointData().GetArray(field):
                self.logger.error("Centerlines do not contain the '%s' data field." % field)
                return False

        return True

    def get_cell_data(self, field):
        """ Get the data for the given cell field names.
        """
        cell_data = self.centerline.GetCellData().GetArray(field)
        return v2n(cell_data)

    def get_point_data(self, field):
        """ Get the data for the given cell field names.
        """
        point_data = self.centerline.GetPointData().GetArray(field)
        return v2n(point_data)

    def discretize(self, params):
        """
        Do the actual discretization
        """
        # step 1: create nodes / areas / segment lengths
        self.discretize_branches(params)

        # step 2: create connectivity
        self.discretize_bifurcations()

    def discretize_branches(self, params):
        """
        Create refined physical quantities: nodes / areas / segment lengths
        """
        # get names of point data fields
        br_name = self.PointDataFields.BRANCH
        p_name = self.PointDataFields.PATH

        # get list of all branches
        branch = self.get_point_data(br_name)
        branch_list = np.unique(branch).tolist()
        if -1 in branch_list:
            branch_list.remove(-1)

        # initialize properties of sub-segments in fine mesh to replace those of original mesh
        self.num_seg = 0
        self.cell_data = {'id': {}, br_name: {}, 'area': {}, 'length': {}, 'stenosis': {}, '1d_seg': {}, 'curv': {}}
        self.point_data = {'id': {}, br_name: {}, 'coord': {}, p_name: {}}

        # global counters for unique point/element-ids
        cell_ids_offset = 0
        point_ids_offset = 0

        # loop all branches
        for br in branch_list:
            # determine sampling of branch in interval [0, 1]
            sample_1d, f_sten, area, seg_num_1d = self.get_sampling(br, params)

            # number of samples
            num_seg = sample_1d.shape[0] - 1

            # interpolate branch attributes at defined sampling points
            length, _, _, points, r_curv = self.sample_branch(br, sample_1d)

            # get cell centers to convert point data to cell data
            cell_centers = (sample_1d[1:] + sample_1d[:-1]) / 2.0

            # store attributes
            self.cell_data['id'][br] = np.arange(num_seg) + cell_ids_offset
            self.cell_data['area'][br] = params.Acoef * area
            self.cell_data[br_name][br] = np.ones(num_seg) * br
            self.cell_data['length'][br] = params.lcoef * length * np.diff(sample_1d)
            self.cell_data['curv'][br] = interp1d(cell_centers, sample_1d, r_curv / params.lcoef)
            self.cell_data['stenosis'][br] = f_sten
            self.cell_data['1d_seg'][br] = seg_num_1d
            self.point_data['id'][br] = np.arange(num_seg + 1) + point_ids_offset
            self.point_data[br_name][br] = np.ones(num_seg + 1) * br
            self.point_data['coord'][br] = params.lcoef * points
            self.point_data[p_name][br] = params.lcoef * sample_1d * length
            self.num_seg += num_seg

            point_ids_offset += num_seg + 1
            cell_ids_offset += num_seg

    def get_sampling(self, br, params):
        """
        Get sampling points along branch br in interval [0, 1]
        """
        # initially sample branch attributes at all centerline points
        length, path_1d, area, _, _ = self.sample_branch(br)

        # determine number of segments in this branch
        num_seg = int(round(np.sum(length) / params.seg_size))
        if num_seg < params.seg_min_num:
            num_seg = params.seg_min_num

        # get branch ids
        branch = self.get_point_data(self.PointDataFields.BRANCH)

        # get all branch points
        point_ids = np.where(branch == br)

        # get point arrays from centerline
        print("Branch: {}".format(br))
        area_br = self.get_point_data(self.PointDataFields.AREA)[point_ids]
        path_br = self.get_point_data(self.PointDataFields.PATH)[point_ids]

        if params.seg_min_num == -1:
            # take high-fidelity sampling from centerline
            return path_1d, np.zeros(len(path_1d) - 1), np.vstack((area[:-1], area[1:])).T, params.min_num_elems
        else:
            if num_seg > 1:
                if params.seg_size_adaptive:
                    # adaptive mesh based on area change
                    import pwlf
                    sample_1d = pwlf.PiecewiseLinFit(path_1d, area).fitfast(num_seg, pop=3)
                else:
                    # uniform sampling along centerline
                    sample_1d = np.linspace(0, 1, num_seg + 1)

                # get segment lengths
                length_1d = np.diff(sample_1d) * path_br[-1]

                # extract segments
                intervals = []
                for i, sp in enumerate(sample_1d):
                    intervals += [np.argmin(np.abs(path_br - sp * path_br[-1]))]

                # assemble segment properties
                sample_br = [0]
                f_sten_br = []
                a_br = []
                seg_num_1d = []
                for i in range(len(intervals) - 1):
                    # select segment
                    seg = range(intervals[i], intervals[i + 1])
                    if len(seg) > 1:
                        a = area_br[seg]
                        a_br += [[a[0], a[-1]]]

                        # stenosis factor
                        a_0 = np.min(a)
                        a_s = np.max(a)
                        f_sten_br += [1 / a_0 ** 2 * (a_0 / a_s - 1) ** 2]
                        snum = get_1d_seg_num(a, params)
                    else:
                        f_sten_br += [0]
                        a_br += [2 * [area[intervals[i]]]]
                        snum = params.seg_min_num
                    sample_br += [sample_br[-1] + length_1d[i]]
                    seg_num_1d += [snum]

                return np.array(sample_br) / sample_br[-1], np.array(f_sten_br), np.array(a_br), np.array(seg_num_1d)
            else:
                # find stenoses over entire branch (yields between 1 and 3 segments)
                return find_stenoses(area_br, path_br, params)

    def sample_branch(self, br, sample=None):
        """
        Sample points and areas of a centerline group along a path
        Args:
            br: BranchId
            sample: optional, sample points in [0, 1] (default: sample at centerline points)

        Returns:
        path of centerline points in [0, 1],sampled points, sampled areas
        """
        # get point arrays from centerline
        area = self.get_point_data(self.PointDataFields.AREA)
        distance = self.get_point_data(self.PointDataFields.PATH)
        branch = self.get_point_data(self.PointDataFields.BRANCH)
        points = v2n(self.centerline.GetPoints().GetData())

        # get all branch points
        point_ids = np.where(branch == br)

        # distance between points
        dist = distance[point_ids]

        # branch length
        length = dist[-1]

        # map path to [0, 1]
        path_1d = dist / dist[-1]

        # get curvature
        curvature = curvature_radius(points)

        if sample is not None:
            # if sampling is given, interpolate along 1d centerline
            f = lambda x: interp1d(sample, path_1d, x[point_ids])
        else:
            # return quantities along original path
            f = lambda x: x[point_ids]

        return length, path_1d, f(area), f(points), f(curvature)

    def get_junction_length(self, br):
        """
        Get the length of the part of a branch that is within the upstream junction
        Args:
            br: BranchId

        Returns:
        path length within junction
        """
        # does not apply to inlet branch
        if br == 0:
            return 0.0

        # get arrays from centerline
        cent = self.get_point_data(self.PointDataFields.CENT)
        branch = self.get_point_data(self.PointDataFields.BRANCH)
        bifurcation = self.get_point_data(self.PointDataFields.BIFURCATION)
        points = v2n(self.centerline.GetPoints().GetData())

        # first point id of branch
        ip = np.where(branch == br)[0][0]

        # id of upstream junction
        jc = bifurcation[ip - 1]

        # pick a centerline that passes through branch (and upstream junction)
        cid = np.where(cent[ip])[0][0]

        # points of centerline within upstream junction
        point_ids = np.where(np.logical_and(bifurcation == jc, cent[:, cid]))[0]

        # calculate path length
        return np.sum(np.linalg.norm(np.diff(points[point_ids], axis=0), axis=1))

    def discretize_bifurcations(self):
        """
        Create refined sub-segment connectivity
        """
        # get mesh connectivity
        mesh = get_connectivity(self.centerline)

        branch_id = self.get_point_data(self.PointDataFields.BRANCH)

        # tangent on centerline = section normals
        tangent = self.get_point_data(self.PointDataFields.NORMAL)

        # cell id list of 1d bifurcations (first is inlet)
        self.seg_connectivity = []

        # point id on bifurcations
        self.seg_rear = []

        # add joints for bifurcations
        for bf, bifurcation in mesh.items():
            # bifurcation inflow cell is at the end of a branch
            assert bifurcation.inflow is not None, 'bifurcation ' + str(bf) + ' has no inlet branch'
            joint = [self.cell_data['id'][bifurcation.inflow][-1]]

            # bifurcation outflow cells are at the beginning of a branch
            for branch in bifurcation.outflow:
                joint += [self.cell_data['id'][branch][0]]

            self.seg_connectivity += [joint]
            self.seg_rear += [self.point_data['id'][bifurcation.inflow][-1]]
        self.num_bf_geo = len(self.seg_connectivity)

        # add joints for cells within branch
        for branch, cells in self.cell_data['id'].items():
            for i in range(cells.shape[0] - 1):
                self.seg_connectivity += [[cells[i], cells[i + 1]]]
                self.seg_rear += [self.point_data['id'][branch][i + 1]]
        self.num_bf_branch = len(self.seg_connectivity)

        # add outlet cells
        self.terminal = []
        for branch in self.get_outlet_branches():
            self.terminal += [self.cell_data['id'][branch][-1]]

        # add junction information
        self.junctions = {}
        for i, seg in enumerate(self.seg_connectivity):
            # only add "real" junctions (not junctions inside branches)
            if len(seg) > 2:
                # find branch of segment
                for br, segments in self.cell_data['id'].items():
                    if seg[0] in segments:
                        break

                junction = {'tangents': [], 'lengths': [], 'areas': [self.cell_data['area'][br][-1][-1]]}

                # inlet tangent vector
                tangent_in = tangent[branch_id == br][-1]
                tangent_in /= np.linalg.norm(tangent_in)
                junction['tangents'] += [tangent_in.tolist()]

                # loop segments attached to junction
                for s in seg[1:]:
                    # find branch of segment
                    for br, segments in self.cell_data['id'].items():
                        if s in segments:
                            break

                    # angle
                    tangent_out = tangent[branch_id == br][0]
                    tangent_out /= np.linalg.norm(tangent_out)
                    junction['tangents'] += [tangent_out.tolist()]
                    junction['areas'] += [self.cell_data['area'][br][0][0]]
                    junction['lengths'] += [self.get_junction_length(br)]
                self.junctions[i] = junction

    def get_outlet_branches(self):
        """
        Get list of branches connected to outlets
        """
        # branch ids
        br_id = self.get_point_data(self.PointDataFields.BRANCH)
        bf_id = self.get_point_data(self.PointDataFields.BIFURCATION)

        # global node id
        gid = self.get_point_data(self.PointDataFields.NODEID)

        # outlet points are only connected to one cell (skip inlet point)
        ids = vtkIdList()
        outlets = []
        for p in range(self.centerline.GetNumberOfPoints()):
            self.centerline.GetPointCells(p, ids)
            if ids.GetNumberOfIds() == 1 and gid[p] != 0:
                assert br_id[p] != -1, 'bifurcation ' + str(bf_id[p]) + ' is connected to an outlet'
                outlets += [br_id[p]]
        return outlets


def find_stenoses(area, path, params):
    """
    Returns stenosis locations and coefficients along a branch
    """
    # find extrema
    i_min = argrelextrema(area, np.less)[0].tolist()
    i_max = argrelextrema(area, np.greater)[0].tolist()

    # remove numerically close values from extrema lists
    i_min_final = []
    i_max_final = []
    for idx in i_min:
        if not np.isclose(area[idx],area[idx-1]) and not np.isclose(area[idx],area[idx+1]):
            i_min_final.append(idx)
    for idx in i_max:
        if not np.isclose(area[idx],area[idx-1]) and not np.isclose(area[idx],area[idx+1]):
            i_max_final.append(idx)

    i_min = np.array(i_min_final)
    i_max = np.array(i_max_final)
    # truncate minima and maxima (don't understand the logic in this??)

    if i_min.shape[0] > 0:
        a_0 = area[i_min]
    else:
        a_0 = np.array([])

    if i_max.shape[0] > 0:
        a_s = area[i_max]
    else:
        a_s = np.array([])

    if len(a_0) > len(a_s) and len(a_s) > 0:
        a_0 = a_0[:-1]
        i_min = i_min[:-1]
    elif len(a_0) < len(a_s) and len(a_0) > 0:
        a_s = a_s[1:]
        i_max = i_max[1:]

    if i_max.shape[0] > i_min.shape[0]:
        len_unit = i_min.shape[0]
    elif i_max.shape[0] < i_min.shape[0]:
        len_unit = i_max.shape[0]
    else:
        len_unit = i_max.shape[0]

    # stenosis factor
    if len_unit > 0:
        f_sten = 1 / a_0[:len_unit] ** 2 * (a_0[:len_unit] / a_s[:len_unit] - 1) ** 2
    else:
        f_sten = np.array([])

    # no stenosis
    sample_seg = np.array([0, 1])
    a_seg = np.array([[area[0], area[-1]]])
    if len(f_sten) == 0:
        return sample_seg, np.array([0]), a_seg, [get_1d_seg_num(a_seg, params)]
    # one stenosis
    elif len(f_sten) == 1:
        return sample_seg, f_sten, a_seg, [get_1d_seg_num(a_seg, params)]
    # two stenoses, lump together
    elif len(f_sten) == 2:
        return sample_seg, np.array([np.sum(f_sten)]), a_seg, [get_1d_seg_num(a_seg, params)]

    # find largest stenosis
    i_sten = np.argmax(f_sten)

    # pick neighbors of stenosis
    if 0 < i_sten < len(f_sten) - 1:
        seg = [i_sten - 1, i_sten + 1]
    # stenosis at the start/end, lump together
    else:
        return sample_seg, np.array([np.sum(f_sten)]), a_seg, [get_1d_seg_num(a_seg, params)]

    # sample path points
    sample_1d = [0]
    area_1d = []
    for s in seg:
        sample_1d += [path[i_min[s]] / path[-1]]
        area_1d += [area[i_min[s]]]
    sample_1d += [1]

    # store stenosis coefficients
    f_sten_br = [np.sum(f_sten[:i_sten]), f_sten[i_sten], np.sum(f_sten[i_sten + 1:])]

    # area for svOneDSolver
    a_br = [[area[0], area_1d[0]],
            [area[i_min[i_sten]], area[i_min[i_sten]]],
            [area_1d[1], area[-1]]]

    seg_1d = []
    for a in a_br:
        seg_1d += [get_1d_seg_num(a, params)]

    return np.array(sample_1d), np.array(f_sten_br), np.array(a_br), seg_1d


def get_1d_seg_num(area, params):
    """
    Get number of segments for 1d solver
    """
    a_diff = np.max(area) - np.min(area)
    num_seg = int(round(np.sqrt(a_diff) / params.element_size))
    return np.max((params.min_num_elems, num_seg))


class Bifurcation(object):
    """
    Simple class to track inlet and outlets of a bifurcation
    """
    def __init__(self):
        self.inflow = None
        self.outflow = []

    def add_inflow(self, i):
        """
        Add inflow branch i (can only be executed once)
        """
        if self.inflow is None:
            self.inflow = i
        elif self.inflow != i:
            raise ValueError('bifurcation already has inflow id ' + repr(self.inflow))

    def add_outflow(self, i):
        """
        Add outflow branch i
        """
        if i not in self.outflow:
            self.outflow += [i]


def get_connectivity(cent):
    """
    Extract the connectivity (which branches are connected to which bifurcation) from a centerline
    """
    # read arrays from centerline
    branch = v2n(cent.GetPointData().GetArray('BranchId'))
    bifurcation = v2n(cent.GetPointData().GetArray('BifurcationId'))
    bifurcation_list = np.unique(bifurcation).tolist()
    bifurcation_list.remove(-1)

    # get centerline connectivity: which branches are attached to which bifurcation?
    connectivity = {}
    for bf in bifurcation_list:
        connectivity[bf] = Bifurcation()

    # loop all cells
    for c in range(cent.GetNumberOfCells()):
        ele = cent.GetCell(c)
        point_ids = np.array([ele.GetPointIds().GetId(i) for i in range(ele.GetPointIds().GetNumberOfIds())])
        br_ids = branch[point_ids].tolist()

        # find cells that are at borders of bifurcations (two unique RegionIds)
        if np.unique(br_ids).shape[0] == 2:
            # should be one branch and one bifurcation
            assert -1 in br_ids, 'No bifurcation in cell'

            # local node ids of branch and bifurcation (0 or 1)
            i_bf_ele = br_ids.index(-1)
            i_br_ele = int(not i_bf_ele)

            # branch and bifurcation id
            bf = bifurcation[point_ids[i_bf_ele]]
            br = branch[point_ids[i_br_ele]]

            assert bf != -1, 'Multiple bifurcations in cell'
            assert br != -1, 'Multiple branches in cell'

            # branch node is upstream in cell?
            if i_br_ele == 0:
                connectivity[bf].add_inflow(br)
            else:
                connectivity[bf].add_outflow(br)

    for bf, bifurcation in connectivity.items():
        assert len(bifurcation.outflow) >= 2, 'bifurcation ' + str(bf) + ' has less then two outlets'

    return connectivity


def curvature_radius(points):
    """
    Calculate curvature radius from list of points with three adjacent points each
    """
    # form 3 point sets
    p1 = points[:-2]
    p2 = points[1:-1]
    p3 = points[2:]

    p1p2 = p1 - p2
    p1p3 = p1 - p3
    p2p3 = p2 - p3

    # area of triangle
    area = np.linalg.norm(np.cross(p1p2, p1p3), axis=1) / 2.0

    # curvature radius
    curvature = (np.linalg.norm(p1p2) * np.linalg.norm(p1p3) * np.linalg.norm(p2p3)) / (4.0 * area)

    # resample to all points
    path_1d = np.append(0, np.cumsum(np.linalg.norm(np.diff(points, axis=0), axis=1)))
    return interp1d(path_1d, path_1d[1:-1], curvature)


def interp1d(x, xp, fp):
    """
    Mimic behavior of scipy.interpolate.interp1d (extend np.interp to multi-dimensional function values)
    Args:
        x: sample points
        xp: data points
        fp: data values (xD)

    Returns:
        y: interpolated values (xD)
    """
    if len(fp.shape) > 1:
        y = np.zeros((x.shape[0], fp.shape[1]))
        for d in range(fp.shape[1]):
            y[:, d] = np.interp(x, xp, fp[:, d])
    else:
        y = np.interp(x, xp, fp)
    return y



import numpy as np
from typing import Tuple
from itertools import combinations
from time import perf_counter
import scipy.spatial as spatial
import tqdm
from .c_allocate import (
    norm, argwhere_nonzeros,
    argwhere_value_double,
    any_value_double,
    duplicate_map,
    _allocate_patch,
    _allocate_angle
)


# [TODO] the allocator needs to be accelerated to handle large point clouds
def allocate(*args: Tuple[np.ndarray, ...], min_patch_size: int = 10, max_patch_size: int = 20,
             overlap: float = 0.2, feature_angle: float = 30) -> list:
    """
    Create a list of patches from a set of points and normals, if provided, from a point cloud.

    Parameters
    ----------
        points : np.ndarray
            A numpy array of shape (n, d) containing the points of the point cloud.
        normals : np.ndarray
            A numpy array of shape (n, d) containing the normals of the point cloud.
        min_patch_size : int
            The minimum number of points in a patch.
        max_patch_size : int
            The maximum number of points in a patch.
        overlap : float
            The maximum percentage of overlap between patches as a value between 0 and 1.
        feature_angle : float
            The maximum angle allowed between point-wise normal vectors
    Returns
    -------
        patches : list
            A list of tuples containing the points and normals of each patch.
    """
    if len(args) == 0:
        print("Error: No data provided.")
        return [None, None]
    elif len(args) == 1:
        points = args[0]
        normals = None
    else:
        points = args[0]
        normals = args[1]
        magnitudes = norm(normals)
        if any_value_double(magnitudes.flatten(), 0.0):
            e = argwhere_value_double(magnitudes.flatten(), 0.0)
            print("Error: Normals with zero magnitude found at indices:\n{}.".format(e))
            print("Deleting points with zero magnitude normals...")
            print("Points:\n{}".format(points[e, :]))
            print("Normals:\n{}".format(normals[e, :]))
            indices = argwhere_nonzeros(magnitudes.flatten())
            points = points[indices, :]
            normals = normals[indices, :]
            magnitudes = norm(normals)
        normals = normals / magnitudes
    unique_points, unique_inverse, unique_counts = np.unique(points, axis=0, return_inverse=True, return_counts=True)
    max_duplicates = np.max(unique_counts)
    duplicates, duplicate_set = duplicate_map(unique_inverse, unique_counts)
    has_duplicates = len(duplicates) > 0
    if normals is None:
        if has_duplicates:
            print("Warning: Duplicate points found with no normals provided.")
            print("Removing duplicate points...")
            print("Done.")
    else:
        if has_duplicates:
            remove_duplicates = []
            for key in duplicates.keys():
                repeated_indices = duplicates[key]
                repeated_normals = normals[repeated_indices, :]
                comb = combinations(list(range(repeated_normals.shape[0])), 2)
                for i, j in comb:
                    dot = np.dot(repeated_normals[i, :], repeated_normals[j, :])
                    dot = np.clip(dot, -1, 1)
                    angle = np.arccos(dot) * (180 / np.pi)
                    if not np.isclose(angle, 0):
                        if angle < feature_angle:
                            feature_angle = angle
                    else:
                        print("Error: Duplicate points with identical normals found.")
                        print('Points:\n{} \nRepeated Normals:\n{}\nAngle:{}\nCombination:{}'.format(points[repeated_indices, :], repeated_normals,angle,tuple([i,j])))
                        print("Removing duplicate point and normal {}...".format(repeated_indices[j]))
                        remove_duplicates.append(repeated_indices[j])
                        duplicate_set.remove(repeated_indices[j])
                        duplicates[key].remove(repeated_indices[j])
            if len(remove_duplicates) > 0:
                idx = np.arange(points.shape[0]).tolist()
                for i in remove_duplicates:
                    idx.remove(i)
                points = points[idx, :]
                normals = normals[idx, :]
    kdtree = spatial.cKDTree(points)
    overlap = np.clip(overlap, 0, 1)
    max_patch_size = np.min([max_patch_size, points.shape[0]])
    min_patch_size = np.min([min_patch_size, max_patch_size//2])
    patch_points = []
    patch_normals = []
    point_list = np.arange(points.shape[0]).tolist()
    point_set = set(point_list)
    remaining_points = []
    progress_bar = tqdm.tqdm(total=len(point_set), desc='Allocating patches', unit='point', leave=False)
    while len(point_set) > 0:
        progress_start = len(point_set)
        point_idx = point_set.pop()
        _, indices = kdtree.query(points[point_idx, :], k=max_patch_size)
        if normals is not None:
            dots = np.dot(normals[point_idx, :], normals[indices, :].T)
            dots = np.clip(dots, -1, 1)
            angles = np.arccos(dots) * (180 / np.pi)
            indices = np.array(_allocate_angle(point_idx, indices, points[indices, :], normals[indices, :],
                                               feature_angle)).astype(np.int64)
            _, unique_indices = np.unique(points[indices, :], axis=0, return_index=True)
            unique_indices = np.sort(unique_indices)
            indices = indices[unique_indices]
        if len(indices) < min_patch_size:
            remaining_points.append(point_idx)
            continue
        else:
            patch_points.append(points[indices, :])
            if normals is not None:
                patch_normals.append(normals[indices, :])
            else:
                patch_normals.append(None)
            point_set = _allocate_patch(indices, overlap, point_set, duplicate_set)
        progress_end = len(point_set)
        progress_bar.update(progress_start - progress_end)
    progress_bar.close()
    progress_bar = tqdm.tqdm(total=len(remaining_points), desc='Allocating remaining patches', unit='point', leave=False)
    while len(remaining_points) > 0:
        progress_start = len(remaining_points)
        point_idx = remaining_points.pop()
        _, indices = kdtree.query(points[point_idx, :], k=max_patch_size)
        if normals is not None:
            dots = np.dot(normals[point_idx, :], normals[indices, :].T)
            dots = np.clip(dots, -1, 1)
            angles = np.arccos(dots) * (180 / np.pi)
            if np.sum(angles < 180) >= 3:
                indices = np.array(_allocate_angle(point_idx, indices, points[indices, :], normals[indices, :],
                                                   feature_angle)).astype(np.int64)
                _, unique_indices = np.unique(points[indices, :], axis=0, return_index=True)
                unique_indices = np.sort(unique_indices)
                indices = indices[unique_indices]
            else:
                _, unique_indices = np.unique(points[indices, :], axis=0, return_index=True)
                unique_indices = np.sort(unique_indices)
                indices = indices[unique_indices]
                indices = indices[:3]
        patch_points.append(points[indices, :])
        if normals is not None:
            patch_normals.append(normals[indices, :])
        else:
            patch_normals.append(None)
        progress_end = len(remaining_points)
        progress_bar.update(progress_start - progress_end)
    return list(zip(patch_points, patch_normals))
